"""Tools to handle seismic trace headers as Numpy structured arrays."""

import logging
import numpy as np

from collections.abc import Iterable
from numpy.lib import recfunctions as rfn
from sys import byteorder

# try:
#     from numba import jit
# except ImportError:
#     # log.warning("Numba not installed. Using non-optimized code.")
#     def jit(*args, **kwargs):
#         """Create dummy decorator."""
#         def decorator(func):
#             return func
#         return decorator


log = logging.getLogger(__name__)

# (SEG-Y) data formats
_DATAFORMAT = {1: {"desc": "4-byte IBM floating-point", "type": "f", "dtype": np.float32},
               2: {"desc": "4-byte two's complement integer", "type": "i", "dtype": np.int32},
               3: {"desc": "2-byte two's complement integer", "type": "h", "dtype": np.int16},
               5: {"desc": "4-byte IEEE floating-point", "type": "f", "dtype": np.float32},
               6: {"desc": "8-byte IEEE floating-point", "type": "d", "dtype": np.float64},
               8: {"desc": "1-byte two's complement integer", "type": "b", "dtype": np.int8},
               9: {"desc": "8-byte two's complement integer", "type": "q", "dtype": np.int64},
               10: {"desc": "4-byte unsigned integer", "type": "I", "dtype": np.uint32},
               11: {"desc": "2-byte unsigned integer", "type": "H", "dtype": np.uint16},
               12: {"desc": "8-byte unsigned integer", "type": "Q", "dtype": np.uint64},
               16: {"desc": "1-byte unsigned integer", "type": "B", "dtype": np.uint8}}

# encodings of different data types
_DATAENCODING = {"b": {"dtype": "int8", "size": 1},
                 "B": {"dtype": "uint8", "size": 1},
                 "h": {"dtype": "int16", "size": 2},
                 "H": {"dtype": "uint16", "size": 2},
                 "i": {"dtype": "int32", "size": 4},
                 "I": {"dtype": "uint32", "size": 4},
                 "q": {"dtype": "int64", "size": 8},
                 "Q": {"dtype": "uint64", "size": 8},
                 "f": {"dtype": "float32", "size": 4},
                 "d": {"dtype": "float64", "size": 8}}

_SEG2DATAFORMAT = {1: {"desc": "2-byte two's complement integer", "type": "h", "dtype": np.int16},
                   2: {"desc": "4-byte two's complement integer", "type": "i", "dtype": np.int32},
                   3: {"desc": "20-bit floating-point (SEG-D)", "type": "i2", "dtype": np.int32},
                   4: {"desc": "4-byte IEEE floating-point", "type": "f", "dtype": np.float32},
                   5: {"desc": "8-byte IEEE floating-point", "type": "d", "dtype": np.float64}}
# untested: format 3 = 20-bit floating point (SEG-D), stored as scaled 32-bit integer

# Note: the order for SEG2 file header entries is important
_SEG2FILEDESCSTR = ["ACQUISITION_DATE", "ACQUISITION_TIME", "CLIENT", "COMPANY",
                    "GENERAL_CONSTANT", "INSTRUMENT", "JOB_ID", "OBSERVER",
                    "PROCESSING_DATE", "PROCESSING_TIME", "TRACE_SORT", "UNITS", "NOTE"]

# Note: the order for SEG2 trace header entries is important
_SEG2TRACEDESCSTR = ["ALIAS_FILTER", "AMPLITUDE_RECOVERY", "BAND_REJECT_FILTER",
                     "CDP_NUMBER", "CDP_TRACE", "CHANNEL_NUMBER", "DATUM", "DELAY",
                     "DESCALING_FACTOR", "DIGITAL_BAND_REJECT_FILTER",
                     "DIGITAL_HIGH_CUT_FILTER", "DIGITAL_LOW_CUT_FILTER",
                     "END_OF_GROUP", "FIXED_GAIN", "HIGH_CUT_FILTER", "LINE_ID",
                     "LOW_CUT_FILTER", "NOTCH_FREQUENCY", "POLARITY", "RAW_RECORD",
                     "RECEIVER", "RECEIVER_GEOMETRY", "RECEIVER_LOCATION",
                     "RECEIVER_SPECS", "RECEIVER_STATION_NUMBER", "SAMPLE_INTERVAL",
                     "SHOT_SEQUENCE_NUMBER", "SKEW", "SOURCE", "SOURCE_GEOMETRY",
                     "SOURCE_LOCATION", "SOURCE_STATION_NUMBER", "STACK",
                     "STATIC_CORRECTIONS", "TRACE_TYPE", "NOTE"]

_SEG2TRACEDESCALIAS = ["Anti-aliasing filter specs", "Amplitude recovery method",
                       "Acquisition band-rejection filter specs", "CDP number",
                       "Trace number within CDP", "Channel number", "Datum (elevation)",
                       "Delay recording time", "Descaling factor",
                       "Processing digital band-rejection filter specs",
                       "Processing digital high-cut filter specs",
                       "Processing digital low-cut filter specs", "Last trace of group flag",
                       "Recording instrument fixed gain (dB)", "Acquisition high-cut filter specs",
                       "Line identification", "Acquisition low-cut filter specs",
                       "Notch filter frequency", "Polarity", "File name of raw record",
                       "Type of receiver (and number of rec. in group)", "Receiver group geometry",
                       "Receiver (group) location (x, y, z)", "Receiver specs",
                       "Receiver station number", "Sampling interval (s)", "Shot sequence number",
                       "Skew value", "Type of source", "Source (array) geometry",
                       "Source (array) location (x, y, z)", "Source station number",
                       "Stack (no. of summed shots)", "Static correction (src, rec, total)",
                       "Trace type", "Further comments"]


def _check(para):
    """Ensure a list exists."""
    if para is not None:
        return list(np.atleast_1d(para))
    return []


def _check_if_contiguous(buffer):
    """Check if a buffer contains contiguous numbers with stride 1."""
    if len(buffer) < 2:
        return 0
    grad = np.diff(buffer)
    if grad.min() == 1 and grad.max() == 1:
        return 1
    else:
        return 0


def _foreign_endian():
    """Return foreign endianess."""
    if byteorder == "little":
        return ">"
    else:
        return "<"


def _native_endian():
    """Return native endianess."""
    if byteorder == "little":
        return "<"
    else:
        return ">"


def _need_swap(dtype, endian="<"):
    """Check whether byte-swapping is required, dependent on Numpy dtype."""
    if endian == "=":
        endian = _native_endian()

    if dtype.isnative:
        if byteorder == "little":
            retval = False if endian == "<" else True
        else:
            retval = False if endian == ">" else True
    else:
        if byteorder == "little":
            retval = True if endian == "<" else False
        else:
            retval = True if endian == ">" else False
    return retval


def _parse_hdef(hdict, endian="="):
    """Parse JSON header definition."""
    hkeys = list(hdict.keys())
    hformats = []
    bytepos = 1

    for key in hkeys:
        keytype = hdict[key]["type"]
        keysize = _DATAENCODING[keytype]["size"]
        hformats.append(f"{endian}{keytype}")
        if bytepos != hdict[key]["byte"]:
            raise ValueError("JSON has duplicates, gaps or overlaps in byte positions at "
                             f"mnemonic {key}, bytepos {hdict[key]['byte']}, expected {bytepos}.")
        bytepos += keysize

    htitles = [hdict[k]["desc"] for k in hkeys]

    return hkeys, hformats, htitles


def _create_dtype(names, formats, titles=None):
    """Create Numpy dtype."""
    if titles is None:
        dtype = np.dtype({"names": names, "formats": formats}, align=False)
    else:
        try:
            dtype = np.dtype({"names": names, "formats": formats, "titles": titles}, align=False)
        except ValueError as err:
            log.warning("Creating numpy dtype with titles caused an error, re-trying without titles.")
            log.warning("Error was '%s'", err)
            dtype = np.dtype({"names": names, "formats": formats}, align=False)

    return dtype


def add_mnemonic(headers, names=None, data=None, dtypes=None):
    """
    Add mnemonic(s) to structured array.

    This function can be used to add, for instance, a trace header
    mnemonic to the corresponding Numpy structured array.

    Parameters
    ----------
    headers : Numpy structured array
        The header structure (e.g., trace headers).
    names : str or list of str
        The trace header mnemonics to add.
    data : value or list of values, array or list of arrays, None
        The data with which to fill the new header slots. If None,
        then the entries will be filled with zeros. If a single value
        is given, or a list of values where the length of the list
        corresponds to the number of mnemonics to add, each value will
        be used to initialize the corresponding new header slots. If
        an array is given where the length of the array corresponds to
        the number of traces in the header array, or a list of arrays
        where each individual array's length equals the number of
        traces, then each array will be used to initialize the
        corresponding new header slots.
    dtypes : data type or list of data types
        The data types for the new header mnemonics. If only a single
        data type is given but multiple mnemonics are added, then the
        single data type will be used for all new mnemonics.

    Returns
    -------
    Numpy structured array
        The original header array with the new mnemonics added and
        initialized.
    """
    if names is None:
        raise ValueError("Need at least one mnemonic name to add.")
    if dtypes is None:
        raise ValueError("Need to specify dtypes for the new header mnemonic(s).")

    nt = len(headers)

    if isinstance(names, (tuple, list)):
        keys = names
    else:
        keys = [names, ]
    nk = len(keys)

    dt = np.atleast_1d(dtypes)
    ndt = len(dt)
    if nk != ndt:
        if ndt == 1:
            for i in np.arange(nk-ndt):
                dt = np.append(dt, dt[0])
        else:
            raise ValueError("Parameter dtypes must be a single dtype or a list "
                             "of dtypes that matches the number of new mnemonics.")

    val = np.atleast_1d(data)
    nv = len(val)
    if val is None:
        newv = [np.zeros(nt) for i in np.arange(nk)]
    elif nk != nv:
        if nv == 1:
            if isinstance(val[0], Iterable):
                if len(val[0]) != nt:
                    raise ValueError("Length of data does not match number of traces.")
                newv = [val[0] for i in np.arange(nk)]
            else:
                newv = [val[0]*np.ones(nt) for i in np.arange(nk)]
        else:
            raise ValueError("Number of data entries does not match number of new mnemonics.")
    else:
        newv = []
        for v in val:
            if isinstance(v, Iterable):
                if len(v) != nt:
                    raise ValueError("Length of data does not match number of traces.")
                newv.append(v)
            else:
                newv.append(v*np.ones(nt))

    if not headers.dtype.isnative:
        headers = headers.newbyteorder().byteswap()

    return rfn.append_fields(headers, keys, data=newv, dtypes=dt.tolist(), usemask=False)


def remove_mnemonic(headers, names=None, allzero=False):
    """
    Remove mnemonic(s) from structured array.

    This function can be used to remove, for instance, a trace header
    mnemonic from the corresponding Numpy structured array. If you remove
    the "data" mnemonic, you end up with just the trace headers but no
    data values anymore.

    Parameters
    ----------
    headers : Numpy structured array
        The header structure (e.g., trace headers).
    names : str or list of str (default: None)
        The trace header mnemonics to remove.
    allzero : bool (default: False)
        If True, all mnemonics that contain only zeros will be removed,
        possibly in addition to the mnemonics specified via 'names'.

    Returns
    -------
    Numpy structured array
        The original header array with the specified mnemonics removed.
    """
    remove = set()

    if names is None and not allzero:
        raise ValueError("Need at least one mnemonic to remove, or allzero=True.")

    if isinstance(names, str):
        remove.add(names)
    elif isinstance(names, list):
        remove.update(set(names))
    elif isinstance(names, set):
        remove.update(names)

    if allzero:
        keys = list(headers.dtype.names)
        keys.remove("data")
        for k in keys:
            if headers[k].min() == 0 and headers[k].max() == 0:
                remove.add(k)

    return rfn.drop_fields(headers, remove, usemask=False, asrecarray=False)


def rename_mnemonic(headers, mapping=None):
    """
    Rename mnemonic(s) in structured array.

    This function can be used to rename, for instance, a trace header
    mnemonic in the corresponding Numpy structured array.

    Parameters
    ----------
    headers : Numpy structured array
        The header structure (e.g., trace headers).
    namemap : dict
        Dictionary mapping old name(s) to new name(s).

    Returns
    -------
    Numpy structured array
        The original header array with the specified mnemonics renamed.
    """
    if mapping is None:
        raise ValueError("Need a dictionary to map old names to new names.")

    return rfn.rename_fields(headers, mapping)


def ensemble2cube(ensemble, idef="xline", jdef="iline",
                  is_sorted=False, header_trid="trid"):
    """
    Convert a 2D ensemble to a 3D cube.

    The cube's dimensions are defined by the header mnemonics 'idef' and
    'jdef' (plus the vertical axis, usually time or depth). This function will
    pad traces as required to form a regular cube. This can (and definitely
    will) happen when your data cover an area of non-rectangular shape, or in
    case your data coverage has holes. Keep in mind that for strangely shaped
    areas a lot of padding can occur. It is assumed that there are no
    duplicate trace positions in the input data, i.e., each idef/jdef position
    has only one trace.

    Parameters
    ----------
    ensemble : Numpy structured array
        The 2D ensemble to reshape into a cube.
    idef : str, optional (default: 'xline')
        The header mnemonic present in the ensemble's trace headers that
        remains constant along the i-axis.
    jdef : str, optional (default: 'iline')
        The header mnemonic present in the ensemble's trace headers that
        remains constant along the j-axis.
    is_sorted : bool, optional (default: False)
        If the ensemble is already sorted by order=[idef, jdef], set this
        parameter to True to avoid an additional sort (copy). There is no
        check performed whether the ensemble is sorted correctly.
    header_trid : str, optional (default: 'trid')
        Trace header mnemonic to use in order to flag padded traces.
        If set to None, padded traces won't be flagged, otherwise the trace
        identification is set to 3 ('dummy').

    Returns
    -------
    Numpy structured array
        The data reshaped and possibly padded. If requested, padded traces
        have a trace identification of 3; they contain NaN as data values.
        The cube's dimensions ('idef', 'jdef') will be in ascending order.
    """
    keys = ensemble.dtype.names
    if keys is None:
       raise ValueError("No structured array with trace headers given.")
    keys = list(keys)
    if idef is None or jdef is None:
        raise ValueError("Need 'idef' and 'jdef' parameters to form cube.")
    if idef not in keys:
        raise KeyError(f"Mnemonic '{idef}' not found in ensemble's trace headers.")
    if jdef not in keys:
        raise KeyError(f"Mnemonic '{jdef}' not found in ensemble's trace headers.")
    if header_trid is not None and header_trid not in keys:
        raise KeyError(f"Mnemonic '{header_trid}' not found in ensemble's trace headers.")

    nt, ns = ensemble["data"].shape
    if nt < 3:
        log.warning("Reshaping an ensemble with only %d trace(s) is not meaningful.")
        if nt == 0:
            raise ValueError("Input structured array contains no traces.")

    if is_sorted:
        ens = ensemble.view()
    else:
        ens = np.sort(ensemble, order=[idef, jdef])
    xuniq = np.unique(ens[idef])
    yuniq = np.unique(ens[jdef])
    stepx = 1
    if len(xuniq) > 1:
        stepx = np.min(np.diff(xuniq))
    stepy = 1
    if len(yuniq) > 1:
        stepy = np.min(np.diff(yuniq))

    xrange = np.arange(xuniq[0], xuniq[-1]+stepx/2, stepx, dtype=xuniq.dtype)
    yrange = np.arange(yuniq[0], yuniq[-1]+stepy/2, stepy, dtype=yuniq.dtype)
    nx_req = len(xrange)
    ny_req = len(yrange)

    log.info("Cube dimensions: (%d, %d, %d)", nx_req, ny_req, ns)
    log.info("I defined by: '%s' (%d to %d, increment %d)", idef, xrange[0], xrange[-1], stepx)
    log.info("J defined by: '%s' (%d to %d, increment %d)", jdef, yrange[0], yrange[-1], stepy)

    if nx_req*ny_req == nt:
        return np.reshape(ens, newshape=(nx_req, ny_req))
    else:
        log.info("Ensemble2cube is padding %d trace(s).", nx_req*ny_req-nt)

    cube = np.zeros(shape=(nx_req, ny_req), dtype=ensemble.dtype)
    # pre-fill data values with NaN
    cube["data"] = np.nan
    # set other values if standard mnemonics are available
    if "ns" in keys:
        cube["ns"] = ns
    if "dt" in keys:
        cube["dt"] = ens[0]["dt"]
    if "delrt" in keys:
        cube["delrt"] = ens[0]["delrt"]

    # if required, pre-fill trace identification with value 3
    if header_trid is not None:
        cube[header_trid] = 3

    for ix, x in enumerate(xrange):
        # pre-fill idef and jdef headers
        cube[ix, :][idef] = x
        cube[ix, :][jdef] = yrange
        # find global indices of relevant input data
        ensidx = np.where(ens[idef] == x)[0]
        if len(ensidx) > 0:
            # find y-indices where to put input data
            yidx = (ens[ensidx][jdef]-yrange[0])//stepy
            # assign input data
            cube[ix, yidx] = ens[ensidx].copy()

    return cube

###############################################################################
# Alternative implementation for ensemble2cube
###############################################################################

# @jit(nopython=True)
# def _cube_walker(xrange, yrange, xval, yval):
#     idx = []; xvl = []; yvl = []
#     nt = len(xval)
#     if len(yval) != nt:
#         raise ValueError("xval and yval have different lengths")
#     ct = 0
#     for x in xrange:
#         for y in yrange:
#             check = nt-1 if ct > nt-1 else ct
#             if xval[check] == x and yval[check] == y:
#                 ct += 1
#             else:
#                 idx.append(ct)
#                 xvl.append(x)
#                 yvl.append(y)

#     return idx, xvl, yvl


# def ensemble2cube(ensemble, idef="xline", jdef="iline",
#                   is_sorted=False, header_trid="trid"):
#     """
#     Convert a 2D ensemble to a 3D cube.

#     The cube's dimensions are defined by the header mnemonics 'idef' and
#     'jdef' (plus the vertical axis, usually time or depth). This function will
#     pad traces as required to form a regular cube. This can (and definitely
#     will) happen when your data cover an area of non-rectangular shape, or in
#     case your data coverage has holes. Keep in mind that for strangely shaped
#     areas a lot of padding can occur. It is assumed that there are no
#     duplicate trace positions in the input data, i.e., each idef/jdef position
#     has only one trace.

#     Parameters
#     ----------
#     ensemble : Numpy structured array
#         The 2D ensemble to reshape into a cube.
#     idef : str, optional (default: 'xline')
#         The header mnemonic present in the ensemble's trace headers that
#         remains constant along the i-axis.
#     jdef : str, optional (default: 'iline')
#         The header mnemonic present in the ensemble's trace headers that
#         remains constant along the j-axis.
#     is_sorted : bool, optional (default: False)
#         If the ensemble is already sorted by order=[idef, jdef], set this
#         parameter to True to avoid an additional sort (copy). There is no
#         check performed whether the ensemble is sorted correctly.
#     header_trid : str, optional (default: 'trid')
#         Trace header mnemonic to use in order to flag padded traces.
#         If set to None, padded traces won't be flagged, otherwise the trace
#         identification is set to 3 ('dummy').

#     Returns
#     -------
#     Numpy structured array
#         The data reshaped and possibly padded. If requested, padded traces
#         have a trace identification of 3; they contain NaN as data values.
#         The cube's dimensions ('idef', 'jdef') will be in ascending order.
#     """
#     keys = ensemble.dtype.names
#     if keys is None:
#        raise ValueError("No structured array with trace headers given.")
#     keys = list(keys)
#     if idef is None or jdef is None:
#         raise ValueError("Need 'idef' and 'jdef' parameters to form cube.")
#     if idef not in keys:
#         raise KeyError(f"Mnemonic '{idef}' not found in ensemble's trace headers.")
#     if jdef not in keys:
#         raise KeyError(f"Mnemonic '{jdef}' not found in ensemble's trace headers.")
#     if header_trid is not None and header_trid not in keys:
#         raise KeyError(f"Mnemonic '{header_trid}' not found in ensemble's trace headers.")

#     nt, ns = ensemble["data"].shape
#     if nt < 3:
#         log.warning("Reshaping an ensemble with only %d trace(s) is not meaningful.")
#         if nt == 0:
#             raise ValueError("Input structured array contains no traces.")

#     if is_sorted:
#         cube = ensemble.copy()
#     else:
#         cube = np.sort(ensemble, order=[idef, jdef])
#     xuniq = np.unique(cube[idef])
#     yuniq = np.unique(cube[jdef])
#     stepx = 1
#     if len(xuniq) > 1:
#         stepx = np.min(np.diff(xuniq))
#     stepy = 1
#     if len(yuniq) > 1:
#         stepy = np.min(np.diff(yuniq))

#     xrange = np.arange(xuniq[0], xuniq[-1]+stepx/2, stepx, dtype=xuniq.dtype)
#     yrange = np.arange(yuniq[0], yuniq[-1]+stepy/2, stepy, dtype=yuniq.dtype)
#     nx_req = len(xrange)
#     ny_req = len(yrange)

#     log.info("Cube dimensions: (%d, %d, %d)", nx_req, ny_req, ns)
#     log.info("I defined by: '%s' (%d to %d, increment %d)", idef, xrange[0], xrange[-1], stepx)
#     log.info("J defined by: '%s' (%d to %d, increment %d)", jdef, yrange[0], yrange[-1], stepy)

#     if nx_req*ny_req == nt:
#         return np.reshape(cube, newshape=(nx_req, ny_req))
#     else:
#         log.info("Ensemble2cube is padding %d trace(s).", nx_req*ny_req-nt)

#     idx, xvl, yvl = _cube_walker(xrange, yrange, cube[idef], cube[jdef])

#     assert len(idx) == (nx_req*ny_req-nt)

#     values = np.zeros((len(idx),), dtype=cube.dtype)
#     values[:]["data"] = np.nan
#     values[:][idef] = xvl
#     values[:][jdef] = yvl
#     # set other values if standard mnemonics are available
#     if "ns" in keys:
#         values[:]["ns"] = ns
#     if "dt" in keys:
#         values[:]["dt"] = cube[0]["dt"]
#     if "delrt" in keys:
#         values[:]["delrt"] = cube[0]["delrt"]
#     if header_trid is not None:
#         values[:][header_trid] = 3

#     return np.reshape(np.insert(cube, idx, values), newshape=(nx_req, ny_req))



"""Abstract reader for seismic files."""

import abc
import logging
import mmap
import numpy as np
import pandas as pd
import time

from dataclasses import dataclass
from numpy.lib import recfunctions as rfn

from . import _ibm2ieee
from . import seisio
from . import tools

log = logging.getLogger(__name__)

try:
    from numba import jit
except ImportError:
    # log.warning("Numba not installed. Using non-optimized code.")
    def jit(*args, **kwargs):
        """Create dummy decorator."""
        def decorator(func):
            return func
        return decorator


@jit(["(int64,int64,int64,int64)"], nopython=True)
def _calc_blocks(start, stride, count, block):
    """Calculate parameters for multi block reads."""
    indices = np.empty((count*block,), dtype=np.int64)
    i = 0
    for _ in range(count):
        idx_s = start
        idx_e = idx_s+block
        for idx in range(idx_s, idx_e):
            indices[i] = idx
            i += 1
        start += stride
    return indices


@jit(["(int64,int64)"], nopython=True)
def _create_batches(nt, batch_size):
    """Calculate parameters for batch reads."""
    for ii in np.arange(0, nt, batch_size, dtype=np.int64):
        if ii+batch_size > nt:
            yield ii, nt-ii
        else:
            yield ii, batch_size


class Reader(seisio.SeisIO, abc.ABC):
    """An abstract Reader class for seismic data I/O."""

    @dataclass
    class _IDX():
        """Index-related parameters and objects."""

        grp_by: list = None
        srt_by: list = None
        gord: int = 1
        sord: int = 1
        head: np.ndarray = None
        keys: np.array = None
        ne: int = 0
        nte: np.array = None
        maxnte: int = 0

    @abc.abstractmethod
    def __init__(self, file):
        """Initialize class Reader."""
        super().__init__(file)

        self._idx = self._IDX()

        log.info("Input file: %s", self._fp.file)
        if not self._fp.file.exists():
            raise ValueError(f"File '{self._fp.file}' does not exist.")

    @property
    def vsi(self):
        """
        Get the (vertical) sampling interval.

        Returns
        -------
        int
            Sampling interval, usually in microunits (e.g., microseconds)
        """
        return self._dp.si

    @property
    def delay(self):
        """
        Get the delay (usually delay recording time) of the first trace.

        Returns
        -------
        int
            Delay, usually in milliunits (e.g., milliseconds)
        """
        return self._dp.delay

    @property
    def vaxis(self):
        """
        Get the sampling times or depths.

        Returns
        -------
        Numpy array
            Sampling values of vertical axis, usually in units.
        """
        dt = self._dp.si * 1e-6
        beg = self._dp.delay * 1e-3
        end = beg + (self._dp.ns-1) * dt
        return np.arange(beg, end+dt/2, dt)

    def read_all_headers(self, silent=False):
        """
        Get all trace headers.

        Parameters
        ----------
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace header table.
        """
        if not silent:
            log.info("Reading all %d trace headers from disk...", self._dp.nt)

        st = time.time()
        with open(self._fp.file, "rb") as fio:
            with mmap.mmap(fio.fileno(), length=0, access=mmap.ACCESS_READ, offset=0) as mm:
                h = np.ndarray(shape=(self._dp.nt, ), dtype=self._tr.thdtype, buffer=mm,
                               strides=(self.trsize, ), order='F', offset=self._fp.skip).copy()
        et = time.time()

        if not silent:
            diff = et-st
            if diff < 0.1:
                log.info("Reading all headers took %.3f seconds.", et-st)
            else:
                log.info("Reading all headers took %.1f seconds.", et-st)

        return h

    def read_headers(self, *trcno, silent=False):
        """
        Get one or more trace headers.

        Parameters
        ----------
        *trcno : int(s)
            The trace numbers (zero-based) to read from disk.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace header table.
        """
        trcs = tools._check(trcno)
        nt = len(trcs)
        if nt == 0:
            raise ValueError("No trace numbers requested. Need at least one.")

        if not silent:
            log.info("Reading %d specific trace header(s) from disk...", nt)

        h = np.ndarray((nt, ), dtype=self._tr.thdtype)

        with open(self._fp.file, "rb") as fio:
            for i, trc in enumerate(trcs):
                if trc < 0 or trc >= self._dp.nt:
                    raise ValueError(f"Requested trace no. {trc} out of range [0,{self._dp.nt}).")
                fio.seek(self._fp.skip+trc*self.trsize, 0)
                h[i] = np.fromfile(fio, dtype=self._tr.thdtype, count=1, offset=0)

        return h

    def read_batch_of_headers(self, start=0, nheaders=100, silent=False):
        """
        Get a certain number of trace headers starting at a specific trace.

        Parameters
        ----------
        start : int, optional (default: 0)
            The trace number (zero-based) to start reading from disk.
        nheaders : int, optional (default: 100)
            The number of subsequent traces to read, including 'start' itself.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace header table.
        """
        if nheaders < 1:
            raise ValueError("Parameter 'nheaders' must be greater or equal 1.")
        if start < 0 or start+nheaders-1 >= self._dp.nt:
            raise ValueError(f"Requested batch of headers out of range [0,{self._dp.nt}).")

        if not silent:
            log.info("Reading %d trace header(s) from disk starting at trace %d...",
                     nheaders, start)

        with open(self._fp.file, "rb") as fio:
            with mmap.mmap(fio.fileno(), length=0, access=mmap.ACCESS_READ, offset=0) as mm:
                h = np.ndarray(shape=(nheaders, ), dtype=self._tr.thdtype, buffer=mm,
                               strides=(self.trsize, ),
                               offset=self._fp.skip+start*self.trsize, order='F').copy()

        return h

    def read_multibatch_of_headers(self, start=0, count=None, stride=None,
                                   block=None, silent=False):
        """
        Get multiple batches of trace headers from the seismic file.

        For instance, start=1, count=3, stride=4, block=2 would get you the
        following trace headers: 1, 2, 5, 6, 9, 10 - the start is 1, 2 traces
        are within each block, the stride from the first trace of a block to
        the first trace in the next block is 4 and in total 3 blocks are read.

        For a data set with, for instance, a constant number of traces per
        gather (say, 480) and 500 gathers in total, this function allows you
        to read the first 10 trace headers within each gather using start=0,
        block=10, stride=480, and count=500.

        Parameters
        ----------
        start : int, optional (default: 0)
            The trace number (zero-based) at which to start reading from disk.
        count : int
            The total number of blocks to read.
        stride : int
            The stride between the first traces in each block.
        block : int
            The size of each block.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace header table.
        """
        if start < 0 or start >= self._dp.nt:
            raise ValueError(f"Requested batch of trace headers out of range [0,{self._dp.nt}).")
        if count is None or stride is None or block is None:
            raise ValueError("Need to specify count, stride and block.")

        if not silent:
            log.info("Reading %d block(s) of %d trace header(s) from disk, "
                     "starting at %d with stride %d...", count, block, start, stride)

        indices = _calc_blocks(start, stride, count, block)

        if np.max(indices) >= self._dp.nt:
            raise ValueError("Requested multibatch of trace headers out of "
                             f"range [0,{self._dp.nt}).")
        nheaders = len(indices)

        h = np.ndarray((nheaders, ), dtype=self._tr.thdtype)

        with open(self._fp.file, "rb") as fio:
            for i in np.arange(nheaders):
                fio.seek(self._fp.skip+indices[i]*self.trsize, 0)
                h[i] = np.fromfile(fio, dtype=self._tr.thdtype, count=1, offset=0)

        return h

    def read_dataset(self, silent=False):
        """Get all traces - an alias for read_all_traces()."""
        return self.read_all_traces(silent=silent)

    def read_all_traces(self, silent=False):
        """
        Get all traces (e.g., read the entire file).

        Parameters
        ----------
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace headers and data.
        """
        if not silent:
            log.info("Reading entire file (%d traces) from disk...", self._dp.nt)

        st = time.time()
        with open(self._fp.file, "rb") as fio:
            with mmap.mmap(fio.fileno(), length=0, access=mmap.ACCESS_READ, offset=0) as mm:
                d = np.ndarray(shape=(self._dp.nt, ), dtype=self._tr.trdtype, buffer=mm,
                               offset=self._fp.skip, order='F').copy()
        et = time.time()

        if not silent:
            diff = et-st
            if diff < 0.1:
                log.info("Reading all traces took %.3f seconds.", et-st)
            else:
                log.info("Reading all traces took %.1f seconds.", et-st)

        if self._fp.datfmt == 1:
            if not silent:
                log.info("Converting IBM floats to IEEE floats.")
            data = d["data"].view(f"{self._fp.endian}u4")
            new_data = _ibm2ieee.ibm2ieee32(data, self._fp.endian)
            d["data"] = new_data

        return d

    def read_traces(self, *trcno, silent=False):
        """
        Get one or more traces.

        Parameters
        ----------
        *trcno : int(s)
            The trace numbers (zero-based) to read from disk.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace headers and data
        """
        trcs = tools._check(trcno)
        nt = len(trcs)
        if nt == 0:
            raise ValueError("No trace numbers requested. Need at least one.")

        if not silent:
            log.info("Reading %d specific trace(s) from disk...", nt)

        d = np.ndarray((nt, ), dtype=self._tr.trdtype)

        with open(self._fp.file, "rb") as fio:
            for i, trc in enumerate(trcs):
                if trc < 0 or trc >= self._dp.nt:
                    raise ValueError(f"Requested trace no. {trc} out of range [0,{self._dp.nt}).")
                fio.seek(self._fp.skip+trc*self.trsize, 0)
                d[i] = np.fromfile(fio, dtype=self._tr.trdtype, count=1, offset=0)

        if self._fp.datfmt == 1:
            data = d["data"].view(f"{self._fp.endian}u4")
            new_data = _ibm2ieee.ibm2ieee32(data, self._fp.endian)
            d["data"] = new_data

        return d

    def read_batch_of_traces(self, start=0, ntraces=100, silent=False):
        """
        Get a certain number of traces starting at a specific trace.

        Parameters
        ----------
        start : int, optional (default: 0)
            The trace number (zero-based) to start reading from disk.
        ntraces : int, optional (default: 100)
            The number of subsequent traces to read, including 'start' itself.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace headers and data.
        """
        if ntraces < 1:
            raise ValueError("Parameter 'ntraces' must be greater or equal 1.")
        if start < 0 or start+ntraces-1 >= self._dp.nt:
            raise ValueError(f"Requested batch of traces out of range [0,{self._dp.nt}).")

        if not silent:
            log.info("Reading %d trace(s) from disk starting at trace %d...", ntraces, start)

        with open(self._fp.file, "rb") as fio:
            fio.seek(self._fp.skip+start*self.trsize, 0)
            d = np.fromfile(fio, dtype=self._tr.trdtype, count=ntraces, offset=0)

        if self._fp.datfmt == 1:
            data = d["data"].view(f"{self._fp.endian}u4")
            new_data = _ibm2ieee.ibm2ieee32(data, self._fp.endian)
            d["data"] = new_data

        return d

    def read_multibatch_of_traces(self, start=0, count=None, stride=None,
                                  block=None, silent=False):
        """
        Get multiple batches of traces from the seismic file.

        See method read_multibatch_of_headers() for some examples.

        Parameters
        ----------
        start : int, optional (default: 0)
            The trace number (zero-based) at which to start reading from disk.
        count : int
            The total number of blocks to read.
        stride : int
            The stride between the first traces in each block.
        block : int
            The size of each block.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace headers and data.
        """
        if start < 0 or start >= self._dp.nt:
            raise ValueError(f"Requested multibatch of traces out of range [0,{self._dp.nt}).")
        if count is None or stride is None or block is None:
            raise ValueError("Need to specify count, stride and block.")

        if not silent:
            log.info("Reading %d block(s) of %d trace(s) from disk, "
                     "starting at %d with stride %d...", count, block, start, stride)

        indices = _calc_blocks(start, stride, count, block)
        if np.max(indices) >= self._dp.nt:
            raise ValueError(f"Requested multibatch of traces out of range [0,{self._dp.nt}).")
        ntraces = len(indices)

        d = np.zeros((ntraces, ), dtype=self._tr.trdtype)

        with open(self._fp.file, "rb") as fio:
            for i in np.arange(ntraces):
                fio.seek(self._fp.skip+indices[i]*self.trsize, 0)
                d[i] = np.fromfile(fio, dtype=self._tr.trdtype, count=1, offset=0)

        if self._fp.datfmt == 1:
            data = d["data"].view(f"{self._fp.endian}u4")
            new_data = _ibm2ieee.ibm2ieee32(data, self._fp.endian)
            d["data"] = new_data

        return d

    def batches_of_headers(self, batch_size=100, silent=False):
        """
        Loop through all headers in blocks (using a generator).

        Parameters
        ----------
        batch_size : int, optional (default: 100)
            The batch size, i.e., number of trace headers to read in one go.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Yields
        ------
        Numpy structured array
            Trace headers.
        """
        nt = np.int64(self._dp.nt)
        if batch_size <= 0:
            raise ValueError("Parameter 'batch_size' cannot be zero or negative.")
        bs = np.int64(batch_size)
        for start, ntraces in _create_batches(nt, bs):
            yield self.read_batch_of_headers(start, ntraces, silent=silent)

    def batches(self, batch_size=100, silent=False):
        """
        Loop through all traces in blocks (using a generator).

        Parameters
        ----------
        batch_size : int, optional (default: 100)
            The batch size, i.e., number of traces to read in one go.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Yields
        ------
        Numpy structured array
            Trace headers and data.
        """
        nt = np.int64(self._dp.nt)
        if batch_size <= 0:
            raise ValueError("Parameter 'batch_size' cannot be zero or negative.")
        bs = np.int64(batch_size)
        for start, ntraces in _create_batches(nt, bs):
            yield self.read_batch_of_traces(start, ntraces, silent=silent)

    def traces(self, silent=False):
        """
        Loop through all traces of the file (using a generator).

        Parameters
        ----------
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Yields
        ------
        Numpy structured array
            Trace headers and data.
        """
        counter = 0
        while counter < self._dp.nt:
            yield self.read_traces(counter, silent=silent)
            counter += 1

    def headers(self, silent=False):
        """
        Loop through all headers of the file (using a generator).

        Parameters
        ----------
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Yields
        ------
        Numpy structured array
            Trace headers.
        """
        hcounter = 0
        while hcounter < self._dp.nt:
            yield self.read_headers(hcounter, silent=silent)
            hcounter += 1

    def create_index(self, group_by=None, sort_by=None, group_order=">", sort_order=">",
                     headers=None, filt=None):
        """
        Create a lookup index for the input file to read ensembles.

        In order to form ensembles, i.e., common-midpoint gathers, common-shot
        gathers, common-receiver gathers, etc., with possibly differing number
        of traces, various traces that are not necessarily stored on disk in
        the correct order have to be read and grouped together.

        This method creates a lookup table (index) where groups of traces are
        formed according to user-supplied trace header mnemonics. Each group
        or ensemble can then be sorted by yet another set of user-supplied
        trace header mnemonics. Before grouping takes place, a filter function
        to restrict traces to be considered can be applied. The order in which
        groups are formed as well as the order in which traces are sorted
        within groups can be specified as either ascending or descending.

        Parameters
        ----------
        group_by : string or iterable of strings
            The header mnemonic(s) to form groups of traces (ensembles). At
            least one mnemonic needs to be supplied, and it must be a valid
            trace header key.
        sort_by : string or iterable of strings, optional
            The header mnemonic(s) by which to sort traces within an ensemble.
            If None, then the traces within an ensemble are returned in the
            order they are stored in the file.
        group_order : char, optional (default: ">")
            Sort order for groups, either ">" for ascending or "<" for
            descending.
        sort_order : char, optional (default: ">")
            Sort order within groups. Either ">" for ascending or "<" for
            descending.
        headers : Numpy structured array or None (default: None)
            The trace header table with values for the entire file. If you
            have previously read headers *for all traces* from disk you can
            supply a complete header array here. If none is available, the
            (relevant) headers are read from the disk file.
        filt:
            Filter function to apply before grouping takes place. The filter
            function can refer to all available trace header mnemonics.

        Examples
        --------
        Simple example:
            create index(headers=myheadertable,
                         group_by='cdp',
                         sort_by='offset')
        This will create ensembles where each ensemble has the 'cdp' trace
        header menmonic in common (i.e., CMP gathers). The CMP gathers will
        be sorted in ascending order. The traces within each ensemble will be
        sorted by the 'offset' trace header mnemonic in ascending order.

        Example using a filter function:
            def filt_func(x): return (x['offset'] <= 3000)
            create_index(group_by=["sx", "sy"],
                         sort_by="offset",
                         sort_order="<",
                         filt=filt_func)
        This will create ensembles where each ensemble has the same shot
        coordinates (basically shot gathers) and traces within each ensemble
        are sorted by offset in descending order. Before the groups are formed,
        all traces with offsets larger than 3000 m are removed through the
        application of the filter function. The resulting ensembles will
        therefore not contain any offsets larger than 3000 m.
        """
        self._idx.grp_by = tools._check(group_by)
        self._idx.srt_by = tools._check(sort_by)
        self._idx.gord = 1 if group_order == ">" else -1
        self._idx.sord = 1 if sort_order == ">" else -1

        if headers is None:
            h = self.read_all_headers()
        else:
            h = headers

        # need to store trace index explicitly as filter function could
        # potentially remove entire entries and the buffer slot would no
        # longer match the trace index
        nt = len(h)
        h = tools.add_mnemonic(h, names="index", data=[np.arange(nt)], dtypes=np.uint32)

        if filt is not None:
            log.info("Ensemble lookup index has filter applied.")
            h = h[np.nonzero(filt(h))]

        filt_keys = self._idx.grp_by + ["index"]
        self._idx.head = rfn.repack_fields(h[filt_keys], align=False)
        self._idx.keys = np.sort(np.unique(self._idx.head[self._idx.grp_by]),
                                 order=self._idx.grp_by)[::self._idx.gord]

        log.info("Created lookup index for %s (order '%s').", self._idx.grp_by, group_order)
        log.info("Each ensemble is sorted by %s (order '%s').", self._idx.srt_by, sort_order)
        log.info("Number of ensembles: %d", self.ne)

    @property
    def ensemble_keys(self):
        """
        Get the ensemble keys (identifiers) for the current index.

        Returns
        -------
        Numpy array
            Ensemble keys.
        """
        if self._idx.keys is not None:
            return self._idx.keys
        log.warning("No index available. You need to call create_index() first.")
        return np.array([])

    @property
    def ne(self):
        """
        Get the number of ensembles (groups) for the current index.

        Returns
        -------
        int
            Number of ensembles.
        """
        if self._idx.keys is not None:
            return len(self._idx.keys)
        log.warning("No index available. You need to call create_index() first.")
        return 0

    @property
    def nensembles(self):
        """
        Get the number of ensembles (groups) for the current index.

        Returns
        -------
        int
            Number of ensembles.
        """
        return self.ne

    @property
    def nte(self):
        """
        Get the number of traces per ensemble key in this index.

        Returns
        -------
        Numpy array
            Number of traces in each ensemble.
        """
        if self._idx.keys is not None:
            return np.array([len(np.nonzero(self._idx.head[self._idx.grp_by] == x)[0])
                             for x in self._idx.keys[self._idx.grp_by]])
        log.warning("No index available. You need to call create_index() first.")
        return np.array([0])

    @property
    def maxnte(self):
        """
        Get the maximum number of traces found in all ensembles.

        Returns
        -------
        int
            Size of largest ensemble.
        """
        return np.max(self.nte)

    def _get_eidx(self, key):
        if type(key) == np.void:
            key_cmp = key
        else:
            key_cmp = np.asarray(key, dtype=self._idx.head[self._idx.grp_by].dtype)
        return self._idx.head[self._idx.head[self._idx.grp_by] == key_cmp]["index"]

    def read_ensemble(self, *idx_keys, silent=False):
        """
        Get one or more ensembles (groups of traces) from a seismic file.

        Parameters
        ----------
        *idx_keys : tuple(s)
            The keys used to identify ensembles.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            Trace headers and data.
        """
        if self._idx.head is None:
            raise RuntimeError("No index available. You need to call create_index() first.")

        for i, val in enumerate(idx_keys):
            trc = self._get_eidx(val)
            if i == 0:
                traces_to_read = trc
            else:
                traces_to_read = traces_to_read.union(trc, sort=None)

        if not silent:
            log.info("Reading ensemble(s) '%s'.", idx_keys[0])

        if tools._check_if_contiguous(traces_to_read):
            d = self.read_batch_of_traces(traces_to_read[0], len(traces_to_read), silent=silent)
        else:
            d = self.read_traces(*traces_to_read, silent=silent)

        return np.sort(d, order=self._idx.srt_by)[::self._idx.sord]

    def ensembles(self, silent=False):
        """
        Loop through all ensembles.

        Parameters
        ----------
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Yields
        ------
        Numpy structured array
            Trace headers and data.
        """
        if self._idx.keys is not None:
            for e in self._idx.keys:
                yield self.read_ensemble(e, silent=silent)
        else:
            raise RuntimeError("No index available. You need to call create_index() first.")

    def thstat(self, headers=None, ntmax=None):
        """
        Determine statistics for each trace header mnemonic.

        Parameters
        ----------
        headers : Numpy structured array, optional (default: None)
            The trace header values. If None, then all trace headers will
            be read from disk. If a structured array contains the data as
            well, they will be dropped before calculating the statistics.
        ntmax : int, optional (default: None)
            Maximum number of traces to take into consideration to build
            statistics. Default is None, i.e., all traces are considered.
        """
        log.info("Calculating trace header statistics.")

        if headers is None:
            if ntmax is None:
                h = self.read_all_headers()
            else:
                h = self.read_batch_of_headers(0, ntmax)
        else:
            if headers.dtype.names is not None:
                keys = list(headers.dtype.names)
                if "data" in keys:
                    keys.remove("data")
                h = headers[keys]
            else:
                raise ValueError("No structured array with trace headers given.")

        summary = pd.DataFrame(h).describe().transpose().loc[:, ['min', 'max', 'mean',
                                                                 'std', '25%', '75%']]

        return summary

    def log_thstat(self, thstat=None, traces=None, zero=False, ntmax=None):
        """
        Print statistics for each trace header mnemonic.

        Parameters
        ----------
        thstat : Pandas dataframe, optional (default: None)
            A Pandas dataframe as produced by the 'thstat' function.
            If None, this routine uses the 'headers' argument or, if
            no headers are provided, calls 'thstat' itself.
        traces : Numpy structured array, optional (default: None)
            The seismic traces (trace headers plus data) or the seismic
            trace headers (as provided by the 'read_all_headers' function).
        zero : bool, optional (default: False)
            Do not print entries that have a value of zero (False) or print
            all min/max entries, independent of values (True).
        ntmax : int, optional (default: None)
            Maximum number of traces to take into consideration to build
            statistics. Default is None, i.e., all traces are considered.
            Only relevant if df=None.
        """
        if thstat is None and traces is None:
            df = self.thstat(headers=None, ntmax=ntmax)
        elif thstat is not None and traces is None:
            df = thstat.copy()
        elif thstat is None and traces is not None:
            keys = list(traces.dtype.names)
            if "data" in keys:
                keys.remove("data")
            df = self.thstat(headers=traces[keys], ntmax=ntmax)
        else:
            log.warning("Both 'thstat' and 'traces' provided. Using 'thstat'.")
            df = thstat.copy()

        if zero is False:
            msg = "Summary of trace header statistics (zeros excluded):"
            mask = np.any([df["min"] != 0, df["max"] != 0], axis=0)
            df = df.loc[mask, :]
        else:
            msg = "Summary of trace header statistics (zeros included):"

        try:
            from tabulate import tabulate
            log.info("%s\n%s", msg, tabulate(df, headers="keys", tablefmt="psql"))
        except ImportError:
            log.info("%s", msg)
            log.info("%s\n%s", "-------- BEGIN --------", df.to_markdown())
            log.info("%s", "--------- END ---------")

        return df



"""
I/O of seismic files.

The seisio Python module provides basic input/output operations for seismic
data in typical standard formats such as SEG-Y or SU.

In this module, data and trace headers are handled as Numpy structured arrays,
i.e., the full spectrum of functions coming along with Numpy and related
modules can be used. Simplicity and user-friendliness have priority in this
module and its intended use compared to achieving the best performance or
highest throughput. The classes and methods provided are kept deliberately
simple to get students participating our lectures and exercises going with
Python and seismic data in standard industry formats. The classes are not
meant to offer all functionality required in a commercial production
processing environment. Having said that, the module tries to keep
performance in mind (e.g., using memory-mapped files where possible) and also
offers functionality not readily available in other packages that are
available in the Python ecosystem.

Author & Copyright: Dr. Thomas Hertweck, geophysics@email.de

License: GNU Lesser General Public License, Version 3
         https://www.gnu.org/licenses/lgpl-3.0.html
"""

__version__ = "1.2.1"
__author__ = "Thomas Hertweck"
__copyright__ = "(c) 2025 Thomas Hertweck"
__license__ = "GNU Lesser General Public License, Version 3"

# requires Python v3.9+

import json
import logging
import numpy as np
import pandas as pd
import shutil

from pathlib import Path

from . import seg2
from . import segy
from . import su
from . import tools
from . import _txtheader

log = logging.getLogger(__name__)

_FILE_SUFFIX = [".SGY", ".SEGY", ".SEG-Y", ".SEG_Y",   # SEG-Y
                ".SEG2", ".DAT", ".SG2",               # SEG2
                ".SU"]                                 # SU


def input(file, *args, **kwargs):
    """
    Open a seismic file for reading.

    Parameters
    ----------
    file : str or pathlib.Path
        The name of the file to read.
    filetype : str
        Force a specific file type if file suffix is unknown.

    All other arguments are passed to the underlying class once
    the type of input file is determined.
    """
    filetype = kwargs.pop("filetype", None)

    fpath = Path(file)
    if not fpath.exists():
        raise ValueError(f"Input file {fpath} does not exist.")

    if filetype is None:
        suffix = fpath.suffix.upper()
    else:
        suffix = str("." + filetype).upper()
        if suffix not in _FILE_SUFFIX:
            raise ValueError(f"Unknown value '{filetype}' for argument 'filetype'.")

    if suffix in [".SEGY", ".SGY", ".SEG-Y"]:
        return segy.Reader(file, **kwargs)
    elif suffix in [".SEG2", ".DAT", ".S2", ".SG2"]:
        return seg2.Reader(file, **kwargs)
    elif suffix in [".SU"]:
        return su.Reader(file, **kwargs)
    else:
        raise RuntimeError("Cannot determine file type from file name.")


def output(file, **kwargs):
    """
    Open a seismic file for writing.

    Parameters
    ----------
    file : str or pathlib.Path
        The name of the file to write.
    filetype : str
        Force a specific file type if file suffix is unknown.

    All other arguments are passed to the underlying class once
    the type of output file is determined.
    """
    filetype = kwargs.pop("filetype", None)

    fpath = Path(file)

    if filetype is None:
        suffix = fpath.suffix.upper()
    else:
        suffix = str("." + filetype).upper()
        if suffix not in _FILE_SUFFIX:
            raise ValueError(f"Unknown value '{filetype}' for parameter 'filetype'.")

    if suffix in [".SEGY", ".SGY", ".SEG-Y"]:
        return segy.Writer(file, **kwargs)
    elif suffix in [".SEG2", ".DAT", ".S2", ".SG2"]:
        raise NotImplementedError("SEG2 output not implemented as it is an acq. format.")
    elif suffix in [".SU"]:
        return su.Writer(file, **kwargs)
    else:
        raise RuntimeError("Cannot determine file type from file name.")


def segy_thdef_template(outfile):
    """
    Get a template to be used as SEG-Y trace header definition table.

    The template can be adjusted for trace header mnemonic names, types,
    byte offsets etc. and then used later on to allow for a custom-made
    trace header definition table.

    Parameters
    ----------
    outfile : str
        Target name (or directory) for the JSON.
    """
    shutil.copy(Path(__file__).parent/"json/segy_traceheader.json", outfile)


def segy_bhdef_template(outfile):
    """
    Get a template to be used as SEG-Y binary header definition table.

    The template can be adjusted for binary header mnemonic names, types,
    byte offsets etc. and then used later on to allow for a custom-made
    binary header definition table.

    Parameters
    ----------
    outfile : str
        Target name (or directory) for JSON.
    """
    shutil.copy(Path(__file__).parent/"json/segy_binaryheader.json", outfile)


def segy_thdef1_template(outfile):
    """
    Get a template to be used as SEG-Y trace header ext. 1 definition table.

    The template can be adjusted for header mnemonic names, types, byte
    offsets etc. and then used later on to allow for a custom-made trace
    header extension 1 definition table.

    Parameters
    ----------
    outfile : str
        Target name (or directory) for the JSON.
    """
    shutil.copy(Path(__file__).parent/"json/segy_traceheader_ext1.json", outfile)


def su_thdef_template(outfile):
    """
    Get a template to be used as SU trace header definition table.

    The template can be adjusted for trace header mnemonic names, types,
    byte offsets etc. and then used later on to allow for a custom-made
    trace header definition table.

    Parameters
    ----------
    outfile : str
        Target name (or directory) for the JSON.
    """
    shutil.copy(Path(__file__).parent/"json/su_traceheader.json", outfile)


def segy_txthead_template(major_version=1, minor_version=0, fill=True):
    """
    Get a template for a SEG-Y textual file header.

    Parameters
    ----------
    major_version : int, optional (default: 1)
        Major SEG-Y revision number.
    minor_version : int, optional (default: 0)
        Minor SEG-Y revision number.
    fill : boolean, optional (default: True)
        Fill the individual strings with Cxx convention (True) or not.

    Returns
    -------
    list
        A list of strings, 40 card images (strings), each of 80 characters.
    """
    return _txtheader.TxtHeader().template(major_version=major_version,
                                           minor_version=minor_version,
                                           fill=fill)


def _thstat(traces):
    """
    Determine statistics for each trace header mnemonic.

    Parameters
    ----------
    traces : Numpy structured array
        The seismic traces.
    """
    keys = list(traces.dtype.names)
    if "data" in keys:
        keys.remove("data")
    summary = pd.DataFrame(traces[keys]).describe().transpose().loc[:, ['min', 'max', 'mean',
                                                                        'std', '25%', '75%']]
    return summary


def log_thstat(traces, zero=False):
    """
    Print statistics for each trace header mnemonic.

    Parameters
    ----------
    traces : Numpy structured array
        The seismic traces (trace headers plus data) or the seismic
        trace headers (as provided by the 'read_all_headers' function).
    zero : bool, optional (default: False)
        Do not print entries that have a value of zero (False) or print
        all min/max entries, independent of values (True).
    """
    df = _thstat(traces)

    if zero is False:
        msg = "Summary of trace header statistics (zeros excluded):"
        mask = np.any([df["min"] != 0, df["max"] != 0], axis=0)
        df = df.loc[mask, :]
    else:
        msg = "Summary of trace header statistics (zeros included):"

    try:
        from tabulate import tabulate
        log.info("%s\n%s", msg, tabulate(df, headers="keys", tablefmt="psql"))
    except ImportError:
        log.info("%s", msg)
        log.info("%s\n%s", "-------- BEGIN --------", df.to_markdown())
        log.info("%s", "--------- END ---------")

    return df


def log_sgy_default_thdef():
    """
    Print default SEG-Y trace header definition table.

    Returns
    -------
    list
        A list of the trace header mnemonics.
    """
    default_thdef = Path(__file__).parent/"json/segy_traceheader.json"
    thdict = _log_default_thdef(default_thdef)
    return list(thdict.keys())

def log_su_default_thdef():
    """
    Print default SU trace header definition table.

    Returns
    -------
    list
        A list of the trace header mnemonics.
    """
    default_thdef = Path(__file__).parent/"json/su_traceheader.json"
    thdict = _log_default_thdef(default_thdef)
    return list(thdict.keys())

def check_thdef_validity(file):
    """
    Check the validity of a custom trace header definition table.

    Parameters
    ----------
    file : str
        The name of the trace header definition JSON file to be checked.

    Returns
    -------
    bool
        True if definition file is valid, otherwise False.
    """
    with open(file, "r") as io:
        thdict = json.load(io)
    # try to parse and check thdict
    try:
        nn, ff, tt = tools._parse_hdef(thdict, endian=tools._native_endian())
    except ValueError as err:
        log.info("Trace header definition file %s is invalid.", file)
        log.info("Error: %s", err)
        return False

    # try to create dtype
    try:
        dtp = tools._create_dtype(nn, ff, titles=tt)
    except Exception as err:
        log.info("Trace header definition file %s causes problems when "
                 "creating the custom dtype.", file)
        log.info("Error: %s", err)
        return False

    log.info("Trace header definition file %s is valid. Size: %d bytes.",
             file, dtp.itemsize)
    return True

def _log_default_thdef(thdef):
    """Log default trace header definition."""
    with open(thdef, "r") as io:
        thdict = json.load(io)
    tools._parse_hdef(thdict, endian=tools._native_endian())
    msg = "Trace header definition:"
    df = pd.DataFrame(thdict).T
    df["byte"] = df["byte"]-1
    df.rename(columns={"desc": "description"}, inplace=True)
    try:
        from tabulate import tabulate
        log.info("%s\n%s", msg, tabulate(df, headers="keys", tablefmt="psql"))
    except ImportError:
        log.info("%s", msg)
        log.info("%s\n%s", "-------- BEGIN --------", df.to_markdown())
        log.info("%s", "--------- END ---------")
    return thdict



"""Abstract writer for seismic files."""

import abc
import logging
import numpy as np
import pandas as pd

from . import _ibm2ieee
from . import seisio
from . import tools

log = logging.getLogger(__name__)


class Writer(seisio.SeisIO, abc.ABC):
    """An abstract Writer class for seismic data I/O."""

    @abc.abstractmethod
    def __init__(self, file, mode):
        """Initialize class Writer."""
        super().__init__(file)

        log.info("Output file: %s", self._fp.file)

        self._fp.mode = mode
        if self._fp.mode not in ["w", "a"]:
            raise ValueError(f"Unknown value '{self._fp.mode}' for argument 'mode'.")

        f_exists = self._fp.file.exists()

        if f_exists and self._fp.mode == "w":
            log.warning("File %s already exists, truncating file.", self._fp.file)
            with open(self._fp.file, "wb") as _:
                pass  # truncate file
        elif not f_exists and self._fp.mode == "w":
            with open(self._fp.file, "wb") as _:
                pass  # create file
        elif f_exists and self._fp.mode == "a":
            log.info("File %s already exists, appending to file.", self._fp.file)
            log.warning("Note: there are no consistency cross-checks performed.")
        elif not f_exists and self._fp.mode == "a":
            with open(self._fp.file, "wb") as _:
                pass  # create file
        # subsequent writes can now always use "append" mode

        self._head_written = False
        self._tail_written = False

    def traces_template(self, nt=0, headers_only=False):
        """
        Return a traces or header structured array initialized to zero.

        Parameters
        ----------
        nt : int, optional (default: 0)
            The total number of entries in the array.
        headers_only : bool (deault: False)
            Whether to return a trace header structure only (True) or also
            include the two-dimensional "data" buffer (False).

        Returns
        -------
        Numpy structured array
            An empty header table.
        """
        if headers_only:
            return np.zeros((nt, ), dtype=self._tr.thdtype).squeeze()
        else:
            return np.zeros((nt, ), dtype=self._tr.trdtype).squeeze()

    def headers_template(self, nt=0, pandas=False):
        """
        Return a header table.

        Parameters
        ----------
        nt : int, optional (default: 0)
            The total number of entries in the table (number of traces).
        pandas : bool, optional (default: False)
            Whether to return a header table as Pandas dataframe (True)
            or as standard Numpy structured array (False).

        Returns
        -------
        Numpy structured array or Pandas dataframe
            An empty header table.
        """
        headers = self.traces_template(nt=nt, headers_only=True)
        if not pandas:
            return headers
        else:
            return pd.DataFrame.from_records(headers)

    def _headers_transfer(self, headers, remap=None, silent=False):
        """
        Transfer all header mnemonics of input table to current header table.

        This function is particularly useful if converting between different
        trace header definition tables where perhaps not all mnemonics are
        identical.

        Parameters
        ----------
        headers : Numpy structured array
            The header table (source table) to be transferred to the current
            table definition and the corresponding mnemonics.
        remap : dict
            A dictionary remapping headers from the source table to the output
            target table.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        Numpy structured array
            The new header table (target table).
        """
        if headers is None:
            raise ValueError("No headers given on input.")

        nt = len(headers)
        myheaders = self.headers_template(nt=nt)

        if nt == 0:
            return myheaders

        nn = len(headers.dtype.names)
        transferred = np.ones((nn,), dtype=int)

        # transfer
        for i, mnemonic in enumerate(headers.dtype.names):
            if mnemonic in self.mnemonics:
                myheaders[mnemonic] = headers[mnemonic].copy()
                transferred[i] = 0

        # remap; deliberately handled separately from the transfer
        if remap is not None:
            for i, mnemonic in enumerate(headers.dtype.names):
                newkey = remap.get(mnemonic)
                if newkey is not None:
                    if newkey in self.mnemonics:
                        myheaders[newkey] = headers[mnemonic].copy()
                        if not silent:
                            log.info("Remapping mnemonic '%s' to '%s'.", mnemonic, newkey)
                        transferred[i] = 0
                    elif not silent:
                        log.warning("Cannot remap mnemonic '%s' to '%s', key does not exist in "
                                    "target table.", mnemonic, newkey)

        # check
        idx = np.nonzero(transferred)
        for i in idx[0]:
            log.warning("Mnemonic '%s' dropped (not transferable without remap).",
                        headers.dtype.names[i])

        return myheaders

    def write_traces(self, traces=None, data=None, headers=None, remap=None, silent=False):
        """
        Write seismic traces to disk.

        Either traces or data and headers have to be supplied. The headers can
        be in Numpy structured array format, or in a Pandas dataframe format.

        Trace headers and the data will potentially be remapped, transferred
        to the output header table, converted to the output data type, and
        byteswapped if necessary.

        Parameters
        ----------
        traces : Numpy structured array
            A structured array of seismic traces where the actual data are
            stored as key "data" together with the trace headers. This is the
            format that the seisio reader methods use.
        data : 2D Numpy array
            The pure seismic data as two-dimensional Numpy array. If the data
            are not supplied in the correct output data type, they will be
            converted (attention: potential loss of precision).
        headers : Numpy structured array or Pandas dataframe
            The seismic trace headers corresponding to the data. The length of
            the headers array / dataframe must agree with the length of the
            trace-dimension of the data array.
        remap : dict
            A dictionary remapping headers from the existing table in the
            supplied buffer to the output trace header table to be written to
            disk. The key of the dictionary must exist as header mnemonic in
            the supplied buffer, the value of the key must be a mnemonic
            valid for the selected output trace header table. For example,
            { "cdpt": "tracl" } would remap the header 'cdpt' in the input to
            the 'tracl' header in the outout.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).

        Returns
        -------
        int
            The number of traces written in this call.
        """
        if data is None and headers is None and traces is None:
            raise ValueError("No traces (data, headers) to write to disk.")
        if traces is not None and (data is not None or headers is not None):
            raise ValueError("Need either 'traces' or 'data' plus 'headers'.")
        if (headers is None and data is not None) or (data is None and headers is not None):
            raise ValueError("Need both 'data' and 'headers' to write traces to disk.")

        if not self._head_written:
            raise RuntimeError("File has not been initialized; call 'init' first.")
        if self._tail_written:
            raise RuntimeError("File has already been finalized by 'finalize' function call.")

        if traces is not None:
            nt = len(traces)
            if nt == 0:
                raise ValueError("No traces given, structure is empty.")
            try:
                nt, ns = np.shape(traces["data"])
            except ValueError:
                log.error("Supplied Numpy structure 'traces' has no key called 'data'.")
                raise
            if ns == 0:
                raise ValueError("Traces have zero samples.")
            if ns != self._dp.ns:
                raise ValueError("The traces' number of samples differs from constructor value.")

            keys = list(traces.dtype.names)
            keys.remove("data")
            outheaders = self._headers_transfer(traces[keys], remap=remap, silent=silent)
            data = traces["data"]

        if headers is not None:
            nt_h = len(headers)
            nt, ns = data.shape
            if nt == 0:
                raise ValueError("No traces available, data object is empty.")
            if nt != nt_h:
                raise ValueError("Number of traces in data and headers differ.")
            if ns == 0:
                raise ValueError("Traces have zero samples.")
            if ns != self._dp.ns:
                raise ValueError("The traces' number of samples differs from constructor value.")

            if isinstance(headers, pd.DataFrame):
                outheaders = self._headers_transfer(headers.to_records(index=False),
                                                    remap=remap, silent=silent)
            elif isinstance(headers, np.ndarray):
                outheaders = self._headers_transfer(headers, remap=remap, silent=silent)
            else:
                raise TypeError("Given headers object has unsupported type: %s", type(headers))

        if self._fp.datfmt == 1:
            # IBM float
            outdat = _ibm2ieee.ieee2ibm32(data.astype(np.float32), self._fp.endian)
        else:
            if data.dtype.char != self._fp.dtype.char:
                # cast if necessary
                if not silent:
                    log.info("Casting data from type '%s' to '%s'.",
                             data.dtype.char, self._fp.dtype.char)
                outdat = data.astype(self._fp.dtype, copy=True)
            else:
                outdat = data

            if tools._need_swap(outdat.dtype, self._fp.endian):
                # swap bytes if necessary
                if not silent:
                    log.info("Swapping bytes of output data.")
                outdat = outdat.byteswap(inplace=False)

        if not silent:
            log.info("Writing %d trace(s) to disk...", nt)
        with open(self._fp.file, "ab") as fio:
            for i in np.arange(nt):
                fio.write(outheaders[i].tobytes())
                fio.write(outdat[i, :].tobytes())
            self._fp.filesize = fio.tell()

        self._dp.nt += nt
        return nt



"""Conversion of IBM floats to IEEE floats and vice versa."""

import numpy as np

# IBM/IEEE conversion bit masks
_EXPMASK = 0x7f800000
_SIGNMASK = 0x80000000
_MANTMASK = 0x7fffff


def ibm2ieee32(ibm, endian):
    """
    Convert IBM floating point numbers to IEEE format.

    If the 'ibm2ieee' package is installed (recommended), it will be used;
    this is quite fast. Otherwise a slower Numpy-based conversion is used.

    Note: Local conversion code by Robert Kern, 2011.

    Parameters
    ----------
    ibm : np.uint32
        The IBM float(s) (as uint32) to convert.
    endian : char
        Output endianess: ">" for big endian, "<" for little endian.

    Returns
    -------
    np.float32
        IEEE float or array of IEEE floats.
    """
    try:
        from ibm2ieee import ibm2float32
        return ibm2float32(ibm).astype(f"{endian}f")
    except ImportError:
        sign = np.int32(ibm >> 31 & 0x01)
        exponent = np.int32(ibm >> 24 & 0x7f)
        mantissa = np.float64(ibm & 0x00ffffff)/16777216
        return ((1-2*sign)*mantissa*np.power(16.0, exponent-64)).astype(f"{endian}f")


def ieee2ibm32(ieee, endian):
    """
    Convert IEEE floating point numbers to IBM format.

    Note: Local conversion code by Robert Kern, 2011.

    Parameters
    ----------
    ieee : np.float32
        The IEEE float(s) to convert.
    endian : char
        Output endianess: ">" for big endian, "<" for little endian.

    Returns
    -------
    np.uint32
        IBM float or array of IBM floats as np.uint32.
    """
    # ieee = ieee.astype(np.float32)
    asint = ieee.view('i4')
    signbit = asint & _SIGNMASK
    exponent = ((asint & _EXPMASK) >> 23) - 127
    # The IBM 7-bit exponent is to the base 16 and the mantissa is presumed to
    # be entirely to the right of the radix point. In contrast, the IEEE
    # exponent is to the base 2 and there is an assumed 1-bit to the left of
    # the radix point.
    exp16 = ((exponent+1)//4)
    exp_remainder = (exponent+1) % 4
    exp16 += exp_remainder != 0
    downshift = np.where(exp_remainder, 4-exp_remainder, 0)
    ibm_exponent = np.clip(exp16 + 64, 0, 127)
    expbits = ibm_exponent << 24
    # Add the implicit initial 1-bit to the 23-bit IEEE mantissa to get the
    # 24-bit IBM mantissa. Downshift it by the remainder from the exponent's
    # division by 4. It is allowed to have up to 3 leading 0s.
    ibm_mantissa = ((asint & _MANTMASK) | 0x800000) >> downshift
    # Special-case: 0.0
    ibm_mantissa = np.where(ieee, ibm_mantissa, 0)
    expbits = np.where(ieee, expbits, 0)
    return (signbit | expbits | ibm_mantissa).astype(f"{endian}u4")



"""I/O of seismic files in SEG-Y format."""

import json
import logging
import numpy as np
from pathlib import Path

from . import reader
from . import tools
from . import _txtheader
from . import writer
from . import __version__

log = logging.getLogger(__name__)


_SEGYHEADSIZE = 240
_SEGYBINSIZE = 400
_SEGYTXTSIZE = 3200


class Reader(reader.Reader):
    """Class to deal with input of seismic files in SEG-Y format."""

    def __init__(self, file, **kwargs):
        """
        Initialize class Reader.

        Parameters
        ----------
        file : str or pathlib.Path
            The name of the SEG-Y input file to read.
        format : int, optional (default: None)
            Data format of SEG-Y traces, see SEG-Y standard for details.
            Usually, this is determined automatically from the binary file
            header. This parameter allows to manually override the automatic
            detection. A typical value would be 1 (IBM 32-bit floats) or 5
            (IEEE 32-bit floats).
        endian : char, optional (default: None)
            Endianess of the input file, ">" for big endian, "<" for little
            endian. If endian=None (default), then the endianess of the file
            is determined automatically. This parameter can be used to force
            a specific setting.
        fixed : bool, optional (default: None)
            Flag whether the file actually has a fixed trace length with
            identical number of samples and sampling interval on all traces.
            Usually, this is determined automatically, i.e., this flag is a
            way to override the automatic detection. Note that this class
            will only handle files with fixed trace length.
        thdef : str, optional (default: None)
            The name of the SEG-Y trace header definition JSON file. Defaults
            to the standard SEG-Y trace header definition provided by the
            seisio package.
        bhdef : str, optional (default: None)
            The name of the SEG-Y binary header definition JSON file. Defaults
            to the standard SEG-Y binary header definition provided by the
            seisio package.
        txtenc : str, optional (default: None)
            Encoding of the SEG-Y textual file header. Either "ascii" or
            "ebcdic", or None. If None (default), then the encoding of the
            textual file header is determined automatically.
        thext1 : bool, optional (default: False)
            Flag whether trace header extension 1 is used or not.
        thdef1 : str, optional (default: None)
            The name of the SEG-Y trace header extension 1 definition JSON file.
            Defaults to the standard SEG-Y header extension 1 provided by the
            seisio package.
        nthuser : int, optional (default: 0)
            The number of additional 240-byte user-defined trace headers. If
            set, this number must match the number of trace header definition
            files specified as 'thdefu'. Note that trace header extension 1
            must be present if user-defined trace headers are used, according
            to SEG-Y standard.
        thdefu : str or list of str, optional (default: None)
            The name of the SEG-Y user-defined trace header definition JSON
            file(s) if user-defined trace headers are used.
        ntxtrec : int, optional (default: None)
            The number of additional 3200-byte textual header records that
            follow the SEG-Y binary header. Usually, this is determined
            automatically, i.e., this parameter is a way to override the
            automatic detection.
        ntxtrail : int, optional (default: None)
            The number of additional 3200-byte trailer records that follow
            the actual data traces. Usually, this is determined automatically,
            i.e., this parameter is a way to override the automatic detection.
        bin_dt : str, optional (default: "dt")
            The binary header mnemonic specifying the sampling interval.
        bin_ns : str, optional (default: "ns")
            The binary header mnemonic specifying the number of samples.
        bin_format : str, otional (default: "format")
            The binary header mnemonic specifying the data format.
        bin_fixed : str, otional (default: "fixed")
            The binary header mnemonic specifying the fixed trace length flag.
        bin_iconst : str, optional (default: "iconst")
            The binary header mnemonic specifying the integer constant
            16909060_10.
        bin_segymaj : str, optional (default: "segymaj")
            The binary header mnemonic specifying the SEG-Y major revision
            number.
        bin_segymin : str, optional (default: "segymin")
            The binary header mnemonic specifying the SEG-Y minor revision
            number.
        bin_ntxthead : str, optional (default: "ntxthead")
            The binary header mnemonic specifying the number of textual header
            records.
        bin_ntrailer : str, optional (default: "ntrailer")
            The binary header mnemonic specifying the number of trailer stanza
            records.
        bin_maxtrhead : str, optional (default: "maxtrhead")
            The binary header mnemonic specifying the maximum number of
            additional 240-bytes trace headers.
        bin_byteoff : str, optional (default: "byteoff")
            The binary header mnemonic specifying the byte offset to the first
            data trace.
        bin_ens : str, optional (default: "ens")
            The binary header mnemonic specifying the extended number of
            samples.
        bin_edt : str, optional (default: "edt")
            The binary header mnemonic specifying the sextended ampling
            interval.
        bin_ntfile : str, optional (default: "ntfile")
            The binary header mnemonic specifying the number of traces in the
            file.
        mnemonic_delrt : str, optional (default: "delrt")
            The SEG-Y trace header mnemonic specifying the delay recording
            time.
        """
        super().__init__(file)
        self._sgy = self._SEGY()

        self._fp.endian = kwargs.pop("endian", None)
        self._fp.datfmt = kwargs.pop("format", None)
        self._fp.fixed = kwargs.pop("fixed", None)

        self._par["thdef"] = kwargs.pop("thdef", None)
        self._par["bhdef"] = kwargs.pop("bhdef", None)
        self._par["thext1"] = kwargs.pop("thext1", False)
        self._par["thdef1"] = kwargs.pop("thdef1", None)
        self._par["thdefu"] = kwargs.pop("thdefu", None)
        self._par["txtenc"] = kwargs.pop("txtenc", None)
        self._par["nthuser"] = kwargs.pop("nthuser", 0)
        self._par["ntxtrec"] = kwargs.pop("ntxtrec", None)
        self._par["ntxtrail"] = kwargs.pop("ntxtrail", None)

        self._par["bin_dt"] = kwargs.pop("bin_dt", "dt")
        self._par["bin_ns"] = kwargs.pop("bin_ns", "ns")
        self._par["bin_format"] = kwargs.pop("bin_format", "format")
        self._par["bin_fixed"] = kwargs.pop("bin_fixed", "fixed")
        self._par["bin_iconst"] = kwargs.pop("bin_iconst", "iconst")
        self._par["bin_segymaj"] = kwargs.pop("bin_segymaj", "segymaj")
        self._par["bin_segymin"] = kwargs.pop("bin_segymin", "segymin")
        self._par["bin_ntxthead"] = kwargs.pop("bin_ntxthead", "ntxthead")
        self._par["bin_ntrailer"] = kwargs.pop("bin_ntrailer", "ntrailer")
        self._par["bin_maxtrhead"] = kwargs.pop("bin_maxtrhead", "maxtrhead")
        self._par["bin_byteoff"] = kwargs.pop("bin_byteoff", "byteoff")
        self._par["bin_ens"] = kwargs.pop("bin_ens", "ens")
        self._par["bin_edt"] = kwargs.pop("bin_edt", "edt")
        self._par["bin_ntfile"] = kwargs.pop("bin_ntfile", "ntfile")
        self._par["mnemonic_delrt"] = kwargs.pop("mnemonic_delrt", "delrt")

        if kwargs:
            for key, val in kwargs.items():
                log.warning("Unknown argument '%s' with value '%s' encountered.", key, str(val))

        self._fp.mode = "r"

        if self._fp.endian is not None and self._fp.endian not in ["<", ">"]:
            raise ValueError(f"Unknown value '{self._fp.endian}' for argument 'endian'.")

        if self._fp.fixed is not None and self._fp.fixed is False:
            raise NotImplementedError("Support of variable-length traces "
                                      "not implemented in this class.")

        if self._par["thdef"] is None:
            self._par["thdef"] = Path(__file__).parent/"json/segy_traceheader.json"

        if self._par["bhdef"] is None:
            self._par["bhdef"] = Path(__file__).parent/"json/segy_binaryheader.json"

        with open(self._par["bhdef"], "r") as io:
            self._sgy.bhdict = json.load(io)

        if self._par["thext1"] is not None and not isinstance(self._par["thext1"], bool):
            raise TypeError("Argument 'thext1' has wrong type, should be a boolean.")
        if self._par["ntxtrec"] is not None and self._par["ntxtrec"] < 0:
            raise ValueError("Value for argument 'ntxtrec' cannot be negative.")
        if self._par["ntxtrail"] is not None and self._par["ntxtrail"] < 0:
            raise ValueError("Value for argument 'ntxtrail' cannot be negative.")
        if self._par["nthuser"] is not None and self._par["nthuser"] < 0:
            raise ValueError("Value for argument 'nthuser' cannot be negative.")

        self._sgy.txthead = _txtheader.TxtHeader(encoding=self._par["txtenc"])
        with open(self._fp.file, "rb") as fio:
            self._sgy.txthead.read(fio)
            fio.seek(0, 2)
            self._fp.filesize = fio.tell()
        self._par["txtenc"] = self._sgy.txthead.encoding
        assert self._sgy.txthead.size == _SEGYTXTSIZE

        self._fp.skip = self._sgy.txthead.size

        tmp_k, tmp_f, tmp_t = tools._parse_hdef(self._sgy.bhdict, endian="<")
        tmp_dtype_le = tools._create_dtype(tmp_k, tmp_f, titles=tmp_t)
        tmp_k, tmp_f, tmp_t = tools._parse_hdef(self._sgy.bhdict, endian=">")
        tmp_dtype_be = tools._create_dtype(tmp_k, tmp_f, titles=tmp_t)

        if self._fp.endian is None:
            self._fp.endian = self._guess_endianess(tmp_dtype_le, tmp_dtype_be)
            log.info("Input file endianess looks to be '%s' (best guess).", self._fp.endian)
        else:
            log.info("Input file endianess set to '%s'.", self._fp.endian)

        self._sgy.bhdtype = tmp_dtype_le if self._fp.endian == "<" else tmp_dtype_be
        assert self._sgy.bhdtype.itemsize == _SEGYBINSIZE

        self._get_fileattr()
        self._set_segy_dtypes()
        self._segy_nt_and_delay()

    @property
    def ntxtrec(self):
        """
        Get the number of extended textual header records.

        Returns
        -------
        int
            Number of extended header records.
        """
        return len(self._sgy.txtrec)

    @property
    def ntxtrail(self):
        """
        Get the number of trailer stanza records.

        Returns
        -------
        int
            Number of trailer stanza records.
        """
        return len(self._sgy.txtrail)

    @property
    def nthuser(self):
        """
        Get the number of user-defined trace headers.

        Returns
        -------
        int
            Number of user-defined headers.
        """
        return self._par["nthuser"]

    @property
    def thext1(self):
        """
        Get flag whether trace header extension 1 is present.

        Returns
        -------
        bool
            True if trace header extension 1 is present, otherwise False.
        """
        return self._par["thext1"]

    @property
    def txthead(self):
        """
        Get the primary SEG-Y textual file header.

        Returns
        -------
        list
            SEG-Y textual header as a list of 40 strings with 80 characters.
        """
        return self._sgy.txthead.header

    @property
    def binhead(self):
        """
        Get the SEG-Y binary file header.

        Returns
        -------
        Numpy structured array
            SEG-Y binary file header.
        """
        return self._sgy.binhead

    def log_txthead(self):
        """
        Log the primary textual file header.

        Returns
        -------
        list
            The textual header.
        """
        return self._log_txthead()

    def log_bhdef(self):
        """Log the SEG-Y binary file header definition."""
        self._log_bhdef()

    def log_binhead(self, zero=False):
        """
        Log the SEG-Y binary file header.

        Parameters
        ----------
        zero : boolean, optional (default: False)
            Whether to print binary header entries that are zero or not.

        Returns
        -------
        Numpy structured array
            The SEG-Y binary header.
        """
        return self._log_binhead(zero=zero)

    @property
    def records(self):
        """
        Get the additional SEG-Y textual header records.

        Returns
        -------
        list
            List of decoded 3200-byte-long strings.
        """
        return self.get_records()

    def get_records(self):
        """
        Get the additional SEG-Y textual header records.

        Returns
        -------
        list
            List of decoded 3200-byte-long strings.
        """
        if self.ntxtrec == 0:
            return []
        else:
            return [t.get_header(decode=True) for t in self._sgy.txtrec]

    @property
    def trailers(self):
        """
        Get the SEG-Y decoded trailer records.

        Returns
        -------
        list
            List of 3200-byte-long byte strings.
        """
        return self.get_trailers()

    def get_trailers(self, decode=True):
        """
        Get the SEG-Y trailer records.

        Parameters
        ----------
        decode : boolean, optional (default: True)
            Whether to decode the trailer or not.

        Returns
        -------
        list
            List of 3200-byte-long byte strings.
        """
        if self.ntxtrail == 0:
            return []
        else:
            return [t.get_header(decode=decode) for t in self._sgy.txtrail]

    def get_txthead(self):
        """
        Get the primary SEG-Y textual file header.

        Returns
        -------
        list
            SEG-Y textual header as a list of 40 strings with 80 characters.
        """
        return self._sgy.txthead.header

    def get_binhead(self):
        """
        Get the SEG-Y binary file header.

        Returns
        -------
        Numpy structured array
            SEG-Y binary file header.
        """
        return self._sgy.binhead

    def _guess_endianess(self, bhdtype_le, bhdtype_be):
        """
        Guess endianess of a SEG-Y file.

        Parameters
        ----------
        bhdtype_le : np.dtype
            dtype for binary header, using little endian byte ordering.
        bhdtype_be : np.dtype
            dtype for binary header, using big endian byte ordering.

        Returns
        -------
        char
            Eendianess, either "<" (little), ">" (big).
        """
        from sys import byteorder

        with open(self._fp.file, "rb") as fio:
            fio.seek(self._fp.skip, 0)
            binhead_le = np.fromfile(fio, dtype=bhdtype_le, count=1)
            fio.seek(self._fp.skip, 0)
            binhead_be = np.fromfile(fio, dtype=bhdtype_be, count=1)

        # The integer constant 16909060_10 (01020304_16). This is used to
        # allow unambiguous detection of the byte ordering to expect for this
        # SEG-Y file. For example, if this field reads as 67305985_10
        # (04030201_16) then the bytes in every Binary File Header, Trace
        # Header and Trace Data field must be reversed as they are read,
        # i.e. converting the endian-ness of the fields. If it reads
        # 33620995_10 (02010403_16) then consecutive pairs of bytes need to
        # be swapped in every Binary File Header, Trace Header and Trace Data
        # field. The byte ordering of all other portions (the Extended Textual
        # Header and Data Trailer) of the SEG-Y file is not affected by this
        # field.
        iconst_le = binhead_le[self._par["bin_iconst"]]
        iconst_be = binhead_be[self._par["bin_iconst"]]
        if iconst_be == 67305985 or iconst_le == 67305985:
            # need to swap, iconst should be 16909060
            # if host is big endian, data are small endian and vice versa
            return "<" if byteorder == "big" else ">"

        # check format
        format_le = binhead_le[self._par["bin_format"]]
        format_be = binhead_be[self._par["bin_format"]]
        if (format_be >= 1) and (format_be <= 16) and (format_le > 16):
            # likely big endian
            return ">"
        if (format_le >= 1) and (format_le <= 16) and (format_be > 16):
            # likely little endian
            return "<"

        # check dt and ns
        dt_le = binhead_le[self._par["bin_dt"]]
        dt_be = binhead_be[self._par["bin_dt"]]
        ns_le = binhead_le[self._par["bin_ns"]]
        ns_be = binhead_be[self._par["bin_ns"]]
        if (dt_be <= 10000) and (ns_be <= 20000):
            # likely big endian but less sure
            return ">"
        if (dt_le <= 10000) and (ns_le <= 20000):
            # likely little endian but less sure
            return "<"

        # still unsure; go with tradition of rev 0
        return ">"

    def _get_fileattr(self):
        """Determine certain attributes by analzying binary file header."""
        with open(self._fp.file, "rb") as fio:
            fio.seek(self._fp.skip, 0)
            self._sgy.binhead = np.fromfile(fio, dtype=self._sgy.bhdtype, count=1)

        self._fp.skip += self._sgy.bhdtype.itemsize

        self._segy_revision()
        self._segy_dataformat()
        self._segy_fixedlen()
        self._segy_headerrec()
        self._segy_trailerrec()
        self._segy_addtrhead()
        self._segy_byteoffset()
        self._segy_ns_and_si()

    def _segy_revision(self):
        """
        Determine SEG-Y revision.

        Major SEG-Y Format Revision Number. This is an 8-bit unsigned value.
        Thus for SEG-Y Revision 2.0 this will be recorded as 02_16. This field
        is mandatory for all versions of SEG-Y, although a value of zero
        indicates 'traditional' SEG-Y conforming to the 1975 standard.
        Minor SEG-Y Format Revision Number. This is an 8-bit unsigned value
        with a radix point between the first and second bytes. Thus for SEG-Y
        Revision 2.0, this will be recorded as 00_16. This field is mandatory
        for all versions of SEG-Y.
        """
        self._sgy.major = int(self._sgy.binhead[self._par["bin_segymaj"]])
        self._sgy.minor = int(self._sgy.binhead[self._par["bin_segymin"]])
        if self._sgy.major == 0:
            log.info("SEG-Y revision: original SEG-Y conforming to 1975 standard.")
        else:
            log.info("SEG-Y revision (according to binary header): %d.%d",
                     self._sgy.major, self._sgy.minor)

    def _segy_dataformat(self):
        """
        Determine SEG-Y data format.

        1 = 4-byte IBM floating-point
        2 = 4-byte, two's complement integer
        3 = 2-byte, two's complement integer
        4 = 4-byte fixed-point with gain (obsolete)
        5 = 4-byte IEEE floating-point
        6 = 8-byte IEEE floating-point
        7 = 3-byte two’s complement integer
        8 = 1-byte, two's complement integer
        9 = 8-byte, two's complement integer
        10 = 4-byte, unsigned integer
        11 = 2-byte, unsigned integer
        12 = 8-byte, unsigned integer
        15 = 3-byte, unsigned integer
        16 = 1-byte, unsigned integer
        """
        dformat = int(self._sgy.binhead[self._par["bin_format"]])

        if self._fp.datfmt is None:
            self._fp.datfmt = dformat
        else:
            if dformat != self._fp.datfmt:
                log.warning("User-requested format (%d) and format stored in binary "
                            "header (%d) differ.", self._fp.datfmt, dformat)
                log.warning("Using format specified through manual user override.")
            if self._fp.datfmt not in tools._DATAFORMAT:
                log.error("Unsupported SEG-Y data format '%d' - check for possible endian "
                          "issue or wrong 'format' argument.", self._fp.datfmt)
                raise ValueError(f"SEG-Y data format '{self._fp.datfmt}' not supported.")

        log.info("Data sample format: %s.", tools._DATAFORMAT[self._fp.datfmt]["desc"])

    def _segy_fixedlen(self):
        """
        Determine SEG-Y fixed trace length flag.

        Fixed length trace flag. A value of one indicates that all traces in
        this SEG-Y file are guaranteed to have the same sample interval,
        number of trace header blocks and trace samples, as specified in
        Binary File Header bytes 3217–3218 or 3281–3288, 3517–3518, and
        3221–3222 or 3289–3292. A value of zero indicates that the length of
        the traces in the file may vary and the number of samples in bytes
        115–116 of the Standard SEG-Y Trace Header and, if present, bytes
        137–140 of SEG-Y Trace Header Extension 1 must be examined to
        determine the actual length of each trace. This field is mandatory
        for all versions of SEG-Y, although a value of zero indicates
        'traditional' SEG-Y conforming to the 1975 standard. Irrespective of
        this flag, it is strongly recommended that corect values for the
        number of samples per trace and sample interval appear in the
        appropriate trace Trace Header locations.
        """
        segy_fixed = int(self._sgy.binhead[self._par["bin_fixed"]])

        if segy_fixed == 0:
            log.warning("SEG-Y fixed-trace-length flag is not set in binary header.")
            if self._fp.fixed is not None and self._fp.fixed is True:
                log.warning("Assuming fixed-length traces as requested by user.")
            elif self._fp.fixed is None and self._sgy.major == 0:
                self._fp.fixed = True
                log.warning("As SEG-Y major revision is 0, assuming fixed-length traces.")
            elif self._fp.fixed is None and self._sgy.major > 0:
                log.error("Fixed-trace-length flag not set but SEG-Y major rev. > 0.")
                log.error("Variable trace lengths not supported. Use parameter 'fixed' "
                          "to override if traces are in fact fixed length.")
                raise NotImplementedError("Support of variable-length traces "
                                          "not implemented in this class.")
        elif segy_fixed == 1:
            self._fp.fixed = True
            log.info("SEG-Y fixed-trace-length flag is set in binary header.")
        else:
            log.warning("SEG-Y fixed-trace-length flag set to unknown value (%d).", segy_fixed)
            if self._fp.fixed is True:
                log.warning("Assuming fixed-length traces as requested by user.")

    def _segy_headerrec(self):
        """
        Determine number of SEG-Y header records.

        Number of 3200-byte, Extended Textual File Header records following
        the Binary Header. If bytes 3521–3528 are nonzero, that field
        overrides this one. A value of zero indicates there are no Extended
        Textual File Header records (i.e. this file has no Extended Textual
        File Header(s)). A value of -1 indicates that there are a variable
        number of Extended Textual File Header records and the end of the
        Extended Textual File Header is denoted by an ((SEG: EndText)) stanza
        in the final record (Section 6.2). A positive value indicates that
        there are exactly that many Extended Textual File Header records.
        Note that, although the exact number of Extended Textual File Header
        records may be a useful piece of information, it will not always be
        known at the time the Binary Header is written and it is not
        mandatory that a positive value be recorded here or in bytes
        3521–3528. It is however recommended to record the number of records
        if possible as this makes reading more effective and supports direct
        access to traces on disk files. In the event that this number exceeds
        32767, set this field to –1 and bytes 3521–3528 to 3600+3200*(number
        of Extended Textual File Header records). Add a further 128 if a
        SEG-Y Tape Label is present.
        """
        if self._par["ntxtrec"] is None:
            segy_num_headrec = int(self._sgy.binhead[self._par["bin_ntxthead"]])
        else:
            segy_num_headrec = self._par["ntxtrec"]

        if segy_num_headrec > 0:
            # fixed number of additional header records
            with open(self._fp.file, "rb") as fio:
                for i in range(segy_num_headrec):
                    fio.seek(self._fp.skip+i*self._sgy.txthead.size, 0)
                    self._sgy.txtrec.append(_txtheader.TxtHeader(encoding=None, info="SEG-Y ext. "
                                                                 f"textual record {str(i+1)}"))
                    self._sgy.txtrec[i].read(fio)
        elif segy_num_headrec == -1:
            # variable number of add. header records
            cont_reading = True
            segy_num_headrec = 0
            with open(self._fp.file, "rb") as fio:
                while cont_reading:
                    fio.seek(self._fp.skip+segy_num_headrec*self._sgy.txthead.size, 0)
                    self._sgy.txtrec.append(_txtheader.TxtHeader(encoding=None,
                                                                 info="SEG-Y ext. textual record "
                                                                 f"{str(segy_num_headrec+1)}"))
                    self._sgy.txtrec[segy_num_headrec].read(fio)
                    headrec = self._sgy.txtrec[segy_num_headrec].get_header().casefold().strip(' ')
                    # ensure we have a valid extended header record stanza
                    if headrec[0:2] == "((":
                        segy_num_headrec += 1
                        if "((SEG: EndText))".casefold().strip(' ') in headrec:
                            cont_reading = False
                    else:
                        self._sgy.txtrec.pop()
                        cont_reading = False
                        if segy_num_headrec == 0:
                            log.warning("The SEG-Y binary header indicated a variable number "
                                        "of additional header records.")
                            log.warning("However, no valid header stanzas found.")
                            log.warning("Use argument 'ntxtrec' for manual override if required.")
        elif segy_num_headrec != 0:
            log.warning("Unknown value (%s) encountered while determining header records.",
                        str(segy_num_headrec))

        self._fp.skip += self.ntxtrec*self._sgy.txthead.size

        log.info("Number of additional textual header records: %d.", self.ntxtrec)

    def _segy_trailerrec(self):
        """
        Determine number of trailer stanza records.

        Number of 3200-byte data trailer stanza records following the last
        trace (4 byte signed integer). A value of 0 indicates there are no
        trailer records. A value of -1 indicates an undefined number of
        trailer records (0 or more) following the data. It is, however,
        recommended to record the number of trailer records if possible as
        this makes reading more efficient.
        """
        if self._par["ntxtrail"] is None:
            segy_num_trailer = int(self._sgy.binhead[self._par["bin_ntrailer"]])
        else:
            segy_num_trailer = self._par["ntxtrail"]

        if segy_num_trailer > 0:
            # fixed number of additional trailer records
            with open(self._fp.file, "rb") as fio:
                for count in range(segy_num_trailer):
                    fio.seek(0, 2)
                    self._sgy.txtrail.append(_txtheader.TxtHeader(encoding="ascii", info="SEG-Y "
                                                                  f"trailer {str(count+1)}"))
                    fio.seek(-(segy_num_trailer-count)*self._sgy.txthead.size, 2)
                    self._sgy.txtrail[count].read(fio)
        elif segy_num_trailer == -1:
            # variable number of trailer stanzas
            cont_reading = True
            segy_num_trailer = 0
            txtsize = self._sgy.txthead.size
            with open(self._fp.file, "rb") as fio:
                while cont_reading:
                    fio.seek(-(txtsize+segy_num_trailer*txtsize), 2)
                    self._sgy.txtrail.append(_txtheader.TxtHeader(encoding="ascii",
                                                                  info="SEG-Y trailer "
                                                                  f"{str(segy_num_trailer+1)}"))
                    self._sgy.txtrail[segy_num_trailer].read(fio)
                    stanza = self._sgy.txtrail[segy_num_trailer].get_header().casefold().strip(' ')
                    # ensure we have a valid stanza
                    if "((SEG:User Data".casefold().strip(' ') in stanza[0:18]:
                        segy_num_trailer += 1
                    else:
                        self._sgy.txtrail.pop()
                        cont_reading = False
                        if segy_num_trailer == 0:
                            log.warning("The SEG-Y binary header indicated a variable number "
                                        "of data trailer stanza records.")
                            log.warning("However, no valid trailer stanzas '((SEG:User Data' found.")
                            log.warning("Use argument 'ntrailer' for manual override if required.")
        elif segy_num_trailer != 0:
            log.warning("Unknown value (%s) encountered while determining "
                        "trailer records.", str(segy_num_trailer))
        if segy_num_trailer > 0:
            log.warning("This SEG-Y file has trailer stanza records. If they contain binary data,")
            log.warning("decoding them is impossible without parsing the XML description. The")
            log.warning("trailer records might be returned as raw buffers for the user to decode.")

        log.info("Number of trailer stanza records: %d.", self.ntxtrail)

    def _segy_addtrhead(self):
        """
        Determine number of additional trace headers.

        Maximum number of additional 240 byte trace headers. A value of zero
        indicates there are no additional 240 byte trace headers. The actual
        number for a given trace may be supplied in bytes 157–158 of SEG-Y
        Trace Header Extension 1.
        """
        segy_num_trhead = int(self._sgy.binhead[self._par["bin_maxtrhead"]])
        if segy_num_trhead == 0:
            # no additional trace headers present
            if self._par["thext1"]:
                log.warning("Binary header indicates there are no additional 240-byte "
                            "trace headers.")
                log.warning("However, argument 'thext1' was provided, implying the "
                            "presence of trace header extension 1.")
                log.warning("User-supplied parameter overturns binary header value.")
                # self._par["thext1"] already set
            if self._par["nthuser"] > 0:
                log.warning("Binary header indicates there are no additional 240-byte "
                            "user-defined trace headers.")
                log.warning("However, argument 'nthuser' was provided, implying the "
                            "presence of %d user-defined trace headers.", self._par["nthuser"])
                log.warning("User-supplied parameter overturns binary header value.")
                # if user-defined headers are present, extension 1 must be present, too
                self._par["thext1"] = True
        elif segy_num_trhead == 1:
            # one additional trace header present, must be extension 1
            self._par["thext1"] = True
            if self._par["nthuser"] > 0:
                log.warning("Binary header indicates there is one additional 240-byte "
                            "trace header (extension 1).")
                log.warning("However, argument 'nthuser' was provided, implying the presence "
                            "of additional %d user-defined trace headers.", self._par["nthuser"])
                log.warning("User-supplied parameter overturns binary header value.")
        else:
            # binary header indicated >=2 additonal trace headers
            self._par["thext1"] = True
            self._par["nthuser"] = segy_num_trhead-1

        msg = " " if self._par["thext1"] else " not "
        log.info("SEG-Y trace header extension 1 is%spresent.", msg)
        log.info("Number of user-defined trace headers: %d", self._par["nthuser"])

    def _segy_byteoffset(self):
        """
        Determine byte offset.

        Byte offset of first trace relative to start of file or stream if
        known, otherwise zero. (64-bit unsigned integer value) This byte
        count will include the initial 3600 bytes of the Textual and this
        Binary File Header plus the Extended Textual Header if present.
        When nonzero, this field overrides the byte offset implied by any
        nonnegative number of Extended Textual Header records present in
        bytes 3505–3506.
        """
        segy_byte_offset = int(self._sgy.binhead[self._par["bin_byteoff"]])
        byte_offset = self._fp.skip
        if segy_byte_offset > 0:
            self._fp.skip = segy_byte_offset
            if segy_byte_offset != byte_offset:
                log.warning("Byte offset to first trace as stored in SEG-Y "
                            "binary header differs from calculated offset.")
                log.warning("Using value from header. Header: %d bytes, calculated: "
                            "%d bytes.", segy_byte_offset, byte_offset)

        log.info("Byte offset of first trace relative to start of file: %d bytes.",
                 self._fp.skip)

    def _segy_ns_and_si(self):
        """
        Determine number of samples and sampling interval.

        Number of samples per data trace. Note: The sample interval and number
        of samples in the Binary File Header should be for the primary set of
        seismic data traces in the file.

        Sampling interval. Microseconds (μs) for time data, Hertz (Hz) for
        frequency data, meters (m) or feet (ft) for depth data. Extended sample
        interval, IEEE double precision (64-bit). If nonzero, this overrides
        the sample interval in bytes 3217–3218 with the same units.
        """
        self._dp.ns = int(self._sgy.binhead[self._par["bin_ns"]])
        extended_ns = int(self._sgy.binhead[self._par["bin_ens"]])
        if extended_ns > 0:
            self._dp.ns = extended_ns

        log.info("Number of samples per data trace: %d.", self._dp.ns)

        self._dp.si = int(self._sgy.binhead[self._par["bin_dt"]])
        extended_dt = np.float64(self._sgy.binhead[self._par["bin_edt"]])
        if extended_dt != 0:
            self._dp.si = extended_dt

        log.info("Sampling interval: %s (unit as per SEG-Y standard).", str(self._dp.si))

    def _segy_nt_and_delay(self):
        """Determine number of traces and delay recording time."""
        segy_num_traces = int(self._sgy.binhead[self._par["bin_ntfile"]])
        non_data = self._fp.skip + self.ntxtrail * self._sgy.txthead.size
        num_traces = int((self.fsize-non_data)/self.trsize)

        if segy_num_traces > 0:
            self._dp.nt = segy_num_traces
            if segy_num_traces != num_traces:
                log.warning("Number of traces as stored in binary header differs from "
                            "calculated number of traces. Using value from header.")
                log.warning("Stored in header: %d, calculated: %d.", segy_num_traces, num_traces)
        else:
            self._dp.nt = num_traces
            if (self.fsize-non_data) % self.trsize != 0:
                log.warning("Length mismatch encountered in file %s; trying to continue.",
                            self._fp.file)
                log.warning("Filesize: %d bytes, trace size: %d bytes, headers and trailers:"
                            " %d bytes.", self.fsize, self.trsize, non_data)

        log.info("Number of data traces in file: %d.", self._dp.nt)

        with open(self._fp.file, "rb") as fio:
            fio.seek(self._fp.skip, 0)
            headers = np.fromfile(fio, dtype=self._tr.thdtype, count=1)
            self._dp.delay = headers[self._par["mnemonic_delrt"]][0]

        log.info("Delay (on first trace): %s (unit as per SEG-Y standard).", str(self._dp.delay))


class Writer(writer.Writer):
    """Class to deal with output of seismic files in SEG-Y format."""

    def __init__(self, file, **kwargs):
        """
        Initialize class Writer.

        Parameters
        ----------
        file : str or pathlib.Path
            The name of the SEG-Y output file to write.
        ns : int
            Number of samples per output trace.
        vsi : int
            (Vertical) sampling interval (typically in microunits).
        endian : char, optional (default: ">")
            Endianess of the input file, ">" for big endian, "<" for little
            endian, "=" for native endian.
        format : int, optional (default: 5)
            Data format of SEG-Y traces, see SEG-Y standard for details.
        segymaj : int, optional (default: 1)
            SEG-Y major revision number.
        segymin : int, optional (default: 0)
            SEG-Y minor revision number.
        txtenc : str, optional (default: "ascii")
            Encoding of the SEG-Y textual file header. Either "ascii" or
            "ebcdic".
        thdef : str, optional (default: None)
            The name of the SEG-Y trace header definition JSON file. Defaults
            to the standard SEG-Y trace header definition provided by the
            seisio package.
        bhdef : str, optional (default: None)
            The name of the SEG-Y binary header definition JSON file. Defaults
            to the standard SEG-Y binary header definition provided by the
            seisio package.
        thext1 : bool, optional (default: False)
            Flag whether trace header extension 1 is used or not.
        thdef1 : str, optional (default: None)
            The name of the SEG-Y trace header extension 1 definition JSON file.
            Defaults to the standard SEG-Y header extension 1 provided by the
            seisio package.
        nthuser : int, optional (default: 0)
            The number of additional 240-byte user-defined trace headers. If
            set, this number must match the number of trace header definition
            files specified as 'thdefu'. Note that trace header extension 1
            must be present if user-defined trace headers are used, according
            to SEG-Y standard.
        thdefu : str or list of str, optional (default: None)
            The name of the SEG-Y user-defined trace header definition JSON
            file(s) if user-defined trace headers are used.
        ntxtrec : int, optional (default: 0)
            The number of additional 3200-byte textual header records that
            follow the SEG-Y binary header. Usually, this is determined
            automatically, i.e., this parameter is a way to override the
            automatic detection.
        ntxtrail : int, optional (default: 0)
            The number of additional 3200-byte trailer records that follow
            the actual data traces. Usually, this is determined automatically,
            i.e., this parameter is a way to override the automatic detection.
        bin_dt : str, optional (default: "dt")
            The binary header mnemonic specifying the sampling interval.
        bin_ns : str, optional (default: "ns")
            The binary header mnemonic specifying the number of samples.
        bin_format : str, otional (default: "format")
            The binary header mnemonic specifying the data format.
        bin_fixed : str, otional (default: "fixed")
            The binary header mnemonic specifying the fixed trace length flag.
        bin_iconst : str, optional (default: "iconst")
            The binary header mnemonic specifying the integer constant
            16909060_10.
        bin_segymaj : str, optional (default: "segymaj")
            The binary header mnemonic specifying the SEG-Y major revision
            number.
        bin_segymin : str, optional (default: "segymin")
            The binary header mnemonic specifying the SEG-Y minor revision
            number.
        bin_ntxthead : str, optional (default: "ntxthead")
            The binary header mnemonic specifying the number of textual header
            records.
        bin_ntrailer : str, optional (default: "ntrailer")
            The binary header mnemonic specifying the number of trailer stanza
            records.
        bin_maxtrhead : str, optional (default: "maxtrhead")
            The binary header mnemonic specifying the maximum number of
            additional 240-bytes trace headers.
        bin_byteoff : str, optional (default: "byteoff")
            The binary header mnemonic specifying the byte offset to the first
            data trace.
        bin_ens : str, optional (default: "ens")
            The binary header mnemonic specifying the extended number of
            samples.
        bin_edt : str, optional (default: "edt")
            The binary header mnemonic specifying the sextended ampling
            interval.
        bin_ntfile : str, optional (default: "ntfile")
            The binary header mnemonic specifying the number of traces in the
            file.
        """
        mode = "w"
        super().__init__(file, mode)

        self._sgy = self._SEGY()

        self._fp.mode = mode
        self._fp.fixed = True
        self._fp.endian = kwargs.pop("endian", ">")
        self._fp.datfmt = kwargs.pop("format", 5)
        self._dp.ns = kwargs.pop("ns", None)
        self._dp.si = kwargs.pop("vsi", None)

        self._par["segymaj"] = kwargs.pop("segymaj", 1)
        self._par["segymin"] = kwargs.pop("segymin", 0)
        self._par["thdef"] = kwargs.pop("thdef", None)
        self._par["bhdef"] = kwargs.pop("bhdef", None)
        self._par["thext1"] = kwargs.pop("thext1", False)
        self._par["thdef1"] = kwargs.pop("thdef1", None)
        self._par["thdefu"] = kwargs.pop("thdefu", None)
        self._par["txtenc"] = kwargs.pop("txtenc", "ascii")
        self._par["nthuser"] = kwargs.pop("nthuser", 0)
        self._par["ntxtrec"] = kwargs.pop("ntxtrec", 0)
        self._par["ntxtrail"] = kwargs.pop("ntxtrail", 0)

        self._par["bin_dt"] = kwargs.pop("bin_dt", "dt")
        self._par["bin_ns"] = kwargs.pop("bin_ns", "ns")
        self._par["bin_format"] = kwargs.pop("bin_format", "format")
        self._par["bin_fixed"] = kwargs.pop("bin_fixed", "fixed")
        self._par["bin_iconst"] = kwargs.pop("bin_iconst", "iconst")
        self._par["bin_segymaj"] = kwargs.pop("bin_segymaj", "segymaj")
        self._par["bin_segymin"] = kwargs.pop("bin_segymin", "segymin")
        self._par["bin_ntxthead"] = kwargs.pop("bin_ntxthead", "ntxthead")
        self._par["bin_ntrailer"] = kwargs.pop("bin_ntrailer", "ntrailer")
        self._par["bin_maxtrhead"] = kwargs.pop("bin_maxtrhead", "maxtrhead")
        self._par["bin_byteoff"] = kwargs.pop("bin_byteoff", "byteoff")
        self._par["bin_ens"] = kwargs.pop("bin_ens", "ens")
        self._par["bin_edt"] = kwargs.pop("bin_edt", "edt")
        self._par["bin_ntfile"] = kwargs.pop("bin_ntfile", "ntfile")

        if kwargs:
            for key, val in kwargs.items():
                log.warning("Unknown argument '%s' with value '%s' encountered.", key, str(val))

        self._par_check()
        self._set_segy_dtypes()

        log.info("Output file endianess set to '%s'.", self._fp.endian)
        log.info("Number of additional textual header records: %d.", self.ntxtrec)
        log.info("Byte offset of first trace relative to start of file: %d bytes.",
                 self._fp.skip)
        log.info("Number of trailer stanza records: %d.", self.ntxtrail)
        msg = " " if self._par["thext1"] else " not "
        log.info("SEG-Y trace header extension 1 is%spresent.", msg)
        log.info("Number of user-defined trace headers: %d", self._par["nthuser"])
        log.info("Creating file according to SEG-Y rev. %d.%d.",
                 self._par["segymaj"], self._par["segymin"])

    @property
    def ntxtrec(self):
        """
        Get the number of extended textual header records.

        Returns
        -------
        int
            Number of extended header records.
        """
        return len(self._sgy.txtrec)

    @property
    def ntxtrail(self):
        """
        Get the number of trailer stanza records.

        Returns
        -------
        int
            Number of trailer stanza records.
        """
        return len(self._sgy.txtrail)

    @property
    def nthuser(self):
        """
        Get the number of user-defined trace headers.

        Returns
        -------
        int
            Number of user-defined headers.
        """
        return self._par["nthuser"]

    @property
    def thext1(self):
        """
        Get flag whether trace header extension 1 is present.

        Returns
        -------
        bool
            True if trace header extension 1 is present, otherwise False.
        """
        return self._par["thext1"]

    def _par_check(self):
        """Check parameters."""
        if self._fp.endian not in ["<", ">", "="]:
            raise ValueError(f"Unknown value '{self._fp.endian}' for argument 'endian'.")

        if self._fp.endian == "=":
            self._fp.endian = tools._native_endian()

        if self._fp.datfmt not in tools._DATAFORMAT:
            raise ValueError(f"Unknown or unsupported 'format' value {self._fp.datfmt}.")

        if self._dp.ns is None or self._dp.ns <= 0:
            raise ValueError("Need parameter 'ns' greater than zero.")

        if self._dp.si is None or self._dp.si <= 0:
            raise ValueError("Need parameter 'vsi' greater than zero.")

        log.info("Output number of samples per data trace: %d.", self.ns)
        log.info("Output data sample format: %s.", tools._DATAFORMAT[self._fp.datfmt]["desc"])
        if self._fp.datfmt == 1:
            log.warning("IBM floats (format 1) are deprecated. "
                        "Consider using IEEE floats (format 5).")

        if self._fp.endian == "<" and self._fp.datfmt == 1:
            log.warning("Resetting endian to '>' for format %s.",
                        tools._DATAFORMAT[self._fp.datfmt]["desc"])
            self._fp.endian = ">"

        if self._par["segymaj"] not in [0, 1, 2]:
            raise ValueError("Parameter 'segymaj' needs to be 0, 1 or 2.")
        if self._par["segymin"] not in [0, 1]:
            raise ValueError("Parameter 'segymin' needs to be 0 or 1.")

        self._sgy.txthead = _txtheader.TxtHeader(encoding=self._par["txtenc"])
        self._fp.skip = self._sgy.txthead.size

        if self._par["bhdef"] is None:
            self._par["bhdef"] = Path(__file__).parent/"json/segy_binaryheader.json"

        with open(self._par["bhdef"], "r") as io:
            self._sgy.bhdict = json.load(io)

        k, f, t = tools._parse_hdef(self._sgy.bhdict, endian=self._fp.endian)
        self._sgy.bhdtype = tools._create_dtype(k, f, titles=t)
        assert self._sgy.bhdtype.itemsize == _SEGYBINSIZE

        self._fp.skip += self._sgy.bhdtype.itemsize

        if self._par["thdef"] is None:
            self._par["thdef"] = Path(__file__).parent/"json/segy_traceheader.json"

        if self._par["ntxtrec"] is not None and self._par["ntxtrec"] < 0:
            raise ValueError("Value for argument 'ntxtrec' cannot be negative.")

        for i in np.arange(self._par["ntxtrec"]):
            self._sgy.txtrec.append(_txtheader.TxtHeader(encoding=self._par["txtenc"],
                                                         info="SEG-Y ext. textual "
                                                         f"record {str(i+1)}"))
            self._fp.skip += self._sgy.txtrec[i].size

        if self._par["thext1"] is not None and not isinstance(self._par["thext1"], bool):
            raise TypeError("Argument 'thext1' has wrong type, should be a boolean.")

        if self._par["ntxtrail"] is not None and self._par["ntxtrail"] < 0:
            raise ValueError("Value for argument 'ntxtrail' cannot be negative.")

        for i in np.arange(self._par["ntxtrail"]):
            self._sgy.txtrail.append(_txtheader.TxtHeader(encoding="ascii", info="SEG-Y "
                                                          f"trailer {str(i+1)}"))

        if self._par["nthuser"] is not None and self._par["nthuser"] < 0:
            raise ValueError("Value for argument 'nthuser' cannot be negative.")

        if self._par["nthuser"] > 0 and self._par["thext1"] is False:
            log.warning("User-defined trace headers require header extension 1 to be present.")
            log.warning("Setting parameter 'thext1' to 'True'.")
            self._par["thext1"] = True

        if self._fp.endian == "<" and self._par["segymaj"] in [0, 1]:
            log.warning("Little-endian byte ordering not strictly compliant with SEG-Y "
                        "rev. %d.", self._par["segymaj"])

        if self.ntxtrec > 0 and self._par["segymaj"] == 0:
            log.warning("Additional textual file header records not compliant with SEG-Y "
                        "rev. 0.")
            log.warning("Setting parameter 'segymaj' to '1'.")
            self._par["segymaj"] = 1

        if (self._par["nthuser"] > 0 or self._par["thext1"]) and self._par["segymaj"] in [0, 1]:
            log.warning("Additional trace headers not compliant with SEG-Y rev. 0 or 1.")
            log.warning("Setting parameter 'segymaj' to '2'.")
            self._par["segymaj"] = 2

        if self._par["ntxtrail"] > 0 and self._par["segymaj"] in [0, 1]:
            log.warning("Trailer records not compliant with SEG-Y rev. 0 or 1.")
            log.warning("Setting parameter 'segymaj' to '2'.")
            self._par["segymaj"] = 2

        if self._dp.ns > 65535 and self._par["segymaj"] in [0, 1]:
            log.warning("Number of samples (%d) not compliant with SEG-Y rev. 0 or 1.",
                        self._dp.ns)
            log.warning("Setting parameter 'segymaj' to '2'.")
            self._par["segymaj"] = 2

    @property
    def txthead_template(self):
        """
        Provide a template for a primary SEG-Y textual file header.

        Returns
        -------
        list
            A list of strings, 40 card images (strings), each of 80 characters.
        """
        from . import segy_txthead_template
        return segy_txthead_template(major_version=self._par["segymaj"],
                                     minor_version=self._par["segymin"], fill=True)

    @property
    def txtrec_template(self):
        """
        Provide a template for an additional SEG-Y textual file header.

        Note: The difference to txthead_template(), which provides a primary
        textual file header, is the lack of the 'Cxx' convention at the
        beginning of each card image.

        Returns
        -------
        list
            A list of strings, 40 card images (strings), each of 80 characters.
        """
        from . import segy_txthead_template
        return segy_txthead_template(major_version=self._par["segymaj"],
                                     minor_version=self._par["segymin"], fill=False)

    def log_bhdef(self):
        """Log the SEG-Y binary file header definition."""
        self._log_bhdef()

    @property
    def binhead_template(self):
        """
        Provide a template for a SEG-Y binary file header.

        Returns
        -------
        Numpy structured array
            SEG-Y binary file header.
        """
        self._sgy.binhead = np.zeros((1, ), dtype=self._sgy.bhdtype)

        self._sgy.binhead[self._par["bin_format"]] = self._fp.datfmt
        if self._dp.si.is_integer():
            self._sgy.binhead[self._par["bin_dt"]] = self._dp.si
            self._sgy.binhead[self._par["bin_edt"]] = 0
        else:
            self._sgy.binhead[self._par["bin_dt"]] = 0
            self._sgy.binhead[self._par["bin_edt"]] = self._dp.si

        if self._dp.ns > 65535:
            self._sgy.binhead[self._par["bin_ns"]] = 0
            self._sgy.binhead[self._par["bin_ens"]] = self._dp.ns
        else:
            self._sgy.binhead[self._par["bin_ns"]] = self._dp.ns
            self._sgy.binhead[self._par["bin_ens"]] = 0

        self._sgy.binhead[self._par["bin_segymaj"]] = self._par["segymaj"]
        self._sgy.binhead[self._par["bin_segymin"]] = self._par["segymin"]
        self._sgy.binhead[self._par["bin_fixed"]] = 1
        self._sgy.binhead[self._par["bin_ntxthead"]] = self.ntxtrec
        self._sgy.binhead[self._par["bin_maxtrhead"]] = 0
        if self.thext1:
            self._sgy.binhead[self._par["bin_maxtrhead"]] = self.nthuser+1
        self._sgy.binhead[self._par["bin_ntrailer"]] = self.ntxtrail
        self._sgy.binhead[self._par["bin_byteoff"]] = self._fp.skip
        self._sgy.binhead[self._par["bin_iconst"]] = 16909060

        return self._sgy.binhead

    def log_binhead(self, binhead=None, zero=False):
        """
        Log the SEG-Y binary file header.

        Parameters
        ----------
        binhead : Numpy structured array, optional (default: None)
            The binary header. If 'None', the internally stored binary header
            (if available) is used.
        zero : boolean, optional (default: False)
            Whether to print binary header entries that are zero or not.
        """
        self._log_binhead(binhead=binhead, zero=zero)

    def log_txthead(self, txthead=None, info=None):
        """
        Log the (primary) textual file header.

        txthead : list of strings or string, optional (default: None)
            The textual header. If 'None', the internally stored textual
            header (if available) is used.
        info : str, optional (default: None)
            A verbal description of this textual header.
        """
        self._log_txthead(txthead=txthead, info=info)

    def init(self, **kwargs):
        """
        Initialize output and write all SEG-Y file headers to disk.

        This includes the standard textual header, the binary header, and
        possibly any extended header records. This function has to be called
        before writing any traces to the file.

        Parameters
        ----------
        textual : list of strings or string, optional (default: None)
            Primary textual file header (will be encoded 'ascii' or 'ebcdic').
            List of 40 strings. If a string is less than 80 characters, it will
            be padded up to the length of the card image. If a string is more
            than 80 characters, it will be truncated. If no textual header is
            supplied, a default template will be used. Alternatively, a single
            string of length 3200 bytes can be supplied.
        binary : Numpy structured array, optional (default: None)
            The SEG-Y binary header corresponding to the binary header
            definition file used in the class constructor. If no binary header
            is supplied, a default template with minimally pre-filled
            information based on parameters supplied to the class constructor
            will be used.
        records: string or list of strings, optional (default: None)
            The additional SEG-Y extended textual header(s). Each header record
            must be 3200-bytes long.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).
        """
        if self._head_written:
            log.warning("init() method called several times; call ignored.")
            return

        txth = kwargs.pop("textual", None)
        binh = kwargs.pop("binary", None)
        hrec = kwargs.pop("records", None)
        silent = kwargs.pop("silent", False)

        if kwargs:
            for key, val in kwargs.items():
                log.warning("Unknown argument '%s' with value '%s' encountered.", key, str(val))

        if txth is None:
            txth = self.txthead_template
            txth[0] = f"C01 This SEG-Y file was created by Python seisio version {__version__}."
            txth[0] = txth[0].ljust(_txtheader._SEGY_CARDLEN)
        self._sgy.txthead.header = txth
        # Should the last entries be checked for SEG-Y revision information
        # and 'END TEXTUAL HEADER' string? A lot of SEG-Y files do not provide
        # such entries although strictly speaking required by the standard.

        if binh is None:
            self._sgy.binhead = self.binhead_template
        else:
            self._sgy.binhead = binh.astype(self._sgy.bhdtype)

        if hrec is not None:
            if isinstance(hrec, str):
                if self.ntxtrec > 1:
                    raise ValueError("Received fewer ext. header records than expected. "
                                     "Check 'ntxtrec' parameter in constructor.")
                elif self.ntxtrec == 0:
                    log.warning("Expected no ext. header records. Ignoring 'records' parameter.")
                else:
                    self._sgy.txtrec[0].header = hrec
            else:
                if self.ntxtrec > len(hrec):
                    raise ValueError("Received fewer ext. header records than expected. "
                                     "Check 'ntxtrec' parameter in constructor.")
                elif self.ntxtrec < len(hrec):
                    log.warning("Received more ext. header records than expected (%d)."
                                "Ignoring add. 'records'.", self.ntxtrec)
                    for i in range(self.ntxtrec):
                        self._sgy.txtrec[i].header = hrec[i]
                else:
                    for i, rec in enumerate(hrec):
                        self._sgy.txtrec[i].header = rec
        elif self.ntxtrec > 0:
            raise ValueError(f"Expected {self.ntxtrec} extended header "
                             "record(s), received none.")

        with open(self._fp.file, "ab") as io:
            self._sgy.txthead.write(io)
            io.write(self._sgy.binhead.tobytes())
            for i in range(self.ntxtrec):
                self._sgy.txtrec[i].write(io)
            self._fp.filesize = io.tell()

        self._head_written = True
        if not silent:
            log.info("Wrote textual and binary file headers and %d add. header record(s).",
                     self.ntxtrec)

    def finalize(self, *args, encode=True, silent=False):
        """
        Finalize output (and write all SEG-Y file trailers to disk).

        This function has to be called after writing all traces to the file.
        Handle errors / inconsistencies gracefully to allow writing a file at
        the end of a job even if something seems wrong.

        Parameters
        ----------
        *args : 3200-byte buffer(s)
            The actual 3200-bytes trailer buffers.
        encode : bool, optional (default: True)
            ASCII-encode the trailers. If binary data are to be written as
            part of trailer stanzas, trailers should not be encoded.
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).
        """
        if self.ntxtrail > 0:
            if not isinstance(encode, bool):
                log.warning("Parameter 'encode' should be a boolean. Reset to 'True'.")
                encode = True

            if len(args) != self.ntxtrail:
                log.warning("Number of provided trailer records differs from expected number (%d).",
                            self.ntxtrail)
                if len(args) > self.ntxtrail:
                    log.warning("Some trailer records will be ignored.")
                else:
                    log.warning("Resetting number of trailer records to %d.", len(args))
                    del self._sgy.txtrail[len(args):]

            with open(self._fp.file, "ab") as io:
                for i, string in enumerate(args):
                    self._sgy.txtrail[i].set_header(string, encode=encode)
                    self._sgy.txtrail[i].write(io)

            log.info("Wrote %d trailer records.")

        self._sgy.binhead[self._par["bin_ntfile"]] = self._dp.nt
        self._sgy.binhead[self._par["bin_ntrailer"]] = self.ntxtrail

        log.info("Finalizing output file and re-writing updated binary header.")

        with open(self._fp.file, "r+b") as io:
            io.seek(self._sgy.txthead.size, 0)
            io.write(self._sgy.binhead.tobytes())
            io.seek(0, 2)
            self._fp.filesize = io.tell()

        log.info("Wrote a total of %d trace(s), file size: %d bytes.",
                 self._dp.nt, self._fp.filesize)

        self._tail_written = True



"""A module to handle textual headers of SEG-Y files."""

import logging

log = logging.getLogger(__name__)

_SEGY_CARDLEN = 80
_SEGY_NUMCARD = 40

# SEGY textual header encoding
_TXT_ENCODING = {"cp037": "EBCDIC",
                 "ascii": "ASCII"}


class TxtHeader:
    """Class TxtHeader to handle textual headers and their encodings."""

    def __init__(self, encoding=None, info="SEG-Y textual file header"):
        """
        Initialize class TxtHead.

        Parameters
        ----------
        enconding : str, optional (default: None)
            Encoding of the textual header. Either 'ebcdic', 'ascii' or None.
            If None, the encoding will be determined automatically when
            reading a textual header. For writing headers, the encoding
            defaults to 'ascii' if None on input.
        info : str, optional (default: "SEG-Y textual file header")
            A verbal description of this textual header.
        """
        self.encoding = encoding
        self.info = info
        self.header = None              # encoded, single string
        self._bytes = None              # raw header
        self._list = None               # encoded, list of strings

    @property
    def size(self):
        """Return the size (in bytes) of this (textual) header."""
        return _SEGY_CARDLEN * _SEGY_NUMCARD

    def __len__(self):
        """Return the size (in bytes) of this (textual) header."""
        return self.size

    @property
    def info(self):
        """Return descriptive information about this (textual) header."""
        return self._info

    @info.setter
    def info(self, value):
        """Set the descriptive information for this (textual) header."""
        self._info = "Unknown" if value is None else value

    @property
    def encoding(self):
        """Return the encoding of this (textual) header."""
        if self._encoding is not None:
            return _TXT_ENCODING[self._encoding]
        return None

    @encoding.setter
    def encoding(self, value):
        """Set the encoding of this (textual) header."""
        if value is None:
            self._encoding = None
        elif isinstance(value, str):
            value = value.lower()
            if value not in ["ebcdic", "ascii"]:
                raise ValueError(f"Unknown value '{value}' for parameter 'encoding'.")
            self._encoding = "cp037" if value == "ebcdic" else "ascii"
        else:
            raise TypeError("Unknown type for parameter 'encoding'.")

    def _ascii_or_ebcdic(self):
        """
        Determine whether a buffer contains ASCII or EBCDIC.

        The code for a space character in ASCII is hex 20, there is no character
        in the EBCDIC alphabet that has the same code. However, the code for a
        space character in EBCDIC is hex 40, which in the ASCII alphabet is the
        code for the '@' symbol. The ASCII codes for the digits 0 through 9 are
        hex 30 to hex 39 which have no correspondence with any EBCDIC codes. The
        EBCDIC codes for the digits 0 through 9 are hex f0 to hex f9 which have
        no correspondence with any ASCII codes. Use this information to 'guess'
        whether a buffer is ASCII or EBCDIC.

        Returns
        -------
        str
            Encoding of this header. Either 'cp037' (EBCDIC) or 'ascii' (ASCII).
        """
        encoding = None
        n_ascsp = n_ebcsp = n_ascdig = n_ebcdig = 0

        for cbyte in self._bytes:
            if cbyte == 0x20:
                n_ascsp += 1
            elif cbyte == 0x40:
                n_ebcsp += 1
            else:
                if 0x30 <= cbyte <= 0x39:
                    n_ascdig += 1
                elif 0xf0 <= cbyte <= 0xf9:
                    n_ebcdig += 1

        if n_ascsp > n_ebcsp:
            # very likely ASCII
            encoding = "ascii"
            # Some SEG-Y appear to be written in EBCDIC but there are lots
            # of ASCII space characters and no EBCDIC space characters
            if n_ebcdig > n_ascdig:
                encoding = "cp037"
        elif n_ebcsp > n_ascsp:
            # very likely EBCDIC
            encoding = "cp037"
        else:
            if n_ascdig > n_ebcdig:
                # likely ASCII
                encoding = "ascii"
            elif n_ebcdig > n_ascdig:
                # likely EBCDIC
                encoding = "cp037"
            else:
                # no clue, go with tradition
                encoding = "cp037"

        log.info("%s encoding looks to be '%s' (best guess).", self._info, _TXT_ENCODING[encoding])

        return encoding

    def _to_list(self):
        """Transfer textual header to list of strings."""
        self._list = [self._header[i:i+_SEGY_CARDLEN] for i in
                      range(0, self.size, _SEGY_CARDLEN)]

    @property
    def header(self):
        """
        Return the textual header as list.

        Returns
        -------
        list
            The textual header. An empty list is returned if no decoded header
            is available.
        """
        if self._list is not None:
            return self._list
        return []

    @header.setter
    def header(self, header):
        """Set the encoded textual header."""
        self.set_header(header, encode=True)

    def get_header(self, decode=True):
        """
        Return the textual header as single string.

        Either the decoded string or the raw buffer can be returned. The raw
        buffer should be requested if it contains binary, which could be the
        case for any textual trailers. They need to be decoded by the user
        based on available information about that textual header and cannot
        be done automatically.

        Parameters
        ----------
        decode : bool, optional (default: True)
            If 'True', the decoded header is returned. If set to 'False', the
            raw header is returned.

        Returns
        -------
        str
            Textual header as single 3200-character long string.
        """
        if decode:
            if self._header is not None:
                return self._header
        else:
            if self._bytes is not None:
                return self._bytes
        return str("")

    def set_header(self, header, encode=True, silent=False):
        """
        Set the textual header.

        The header will be truncated or expanded as necessary to have the
        correct length.

        Parameters
        ----------
        header : str or list of str, or None
            3200-byte long string or list of max. 40 strings, each max. 80
            characters long.
        encode : boolean, optional (default: True)
            Convert (encode) header (True; default) or keep as is (False).
        silent : bool, optional (default: False)
            Whether to suppress all standard logging (True) or not (False).
        """
        if header is None:
            self._header = None
            return

        if isinstance(header, list):
            hdr = ""
            count = 0
            for count, line in enumerate(header):
                hdr = ''.join([hdr, line.ljust(_SEGY_CARDLEN)[:_SEGY_CARDLEN]])
            if count < _SEGY_NUMCARD-1:
                hdr.ljust(self.size)
        else:
            hdr = header.ljust(self.size)[:self.size]

        self._header = hdr

        if encode:
            if not silent:
                log.info("%s encoding set to '%s'.", self._info, _TXT_ENCODING[self._encoding])
            self._to_list()
            if self._encoding is None:
                self._encoding = "ascii"
            self._bytes = self._header.encode(self._encoding)
        else:
            self._bytes = self._header

    def read(self, fp):
        """
        Read the textual header from a file.

        Parameters
        ----------
        fp : I/O stream
            The file pointer must be correctly positioned before calling
            this function.
        """
        self._bytes = fp.read(self.size)

        if len(self._bytes) < self.size:
            log.error("Bytes read: %d, bytes expected: %d.",
                      len(self._bytes), self.size)
            raise EOFError(f"Short read of textual header '{self._info}'.")

        if self._encoding is None:
            self._encoding = self._ascii_or_ebcdic()
        else:
            log.info("%s encoding set to '%s'.", self._info, _TXT_ENCODING[self._encoding])

        try:  # try to decode; can fail if, e.g., trailer contains binary
            self._header = self._bytes.decode(self._encoding)
            self._to_list()
        except ValueError:  # decoding failed, store header as is
            self._header = self._bytes
            self._list = None

    def write(self, fp):
        """
        Write the textual header to a file.

        Parameters
        ----------
        fp : I/O stream
            The file pointer must be correctly positioned before calling
            this function.
        """
        fp.write(self._bytes)

    @classmethod
    def template(self, major_version=1, minor_version=0, fill=True):
        """
        Return a textual header template.

        This function returns a list of 40 strings, each string 80 characters
        long starting with Cxx, where xx is the line number. A user can fill
        the textual header strings as required and use it later on to write a
        SEG-Y file to disk. C39 specifies the SEG-Y revision (defaults to 1.0),
        C40 specifies 'END TEXTUAL HEADER' as required by the SEG-Y standard.

        Note that additional header records do not necessarily consist of 40
        strings with 80 characters each. They are treated as 3200-bytes long
        single strings. Additional header records or trailer stanzas should
        not use the 'Cxx' convention for starting lines - this is reserved for
        the SEG-Y primary textual header.

        Parameters
        ----------
        major_version : int, optional (default: 1)
            Major SEG-Y revision number.
        minor_version : int, optional (default: 0)
            Minor SEG-Y revision number.
        fill : boolean, optional (default: True)
            Fill the individual strings with Cxx convention (True) or not.

        Returns
        -------
        list
            A list of strings, 40 card images (strings), each of 80 characters.
        """
        header = []
        if fill:
            for row in range(_SEGY_NUMCARD-2):
                header.append(f"C{row+1:02d}".ljust(_SEGY_CARDLEN))
            header.append(f"C{_SEGY_NUMCARD-1:02d} SEG-Y_REV{major_version}."
                          f"{minor_version}".ljust(_SEGY_CARDLEN))
            header.append(f"C{_SEGY_NUMCARD:02d} END TEXTUAL HEADER".ljust(_SEGY_CARDLEN))
        else:
            for row in range(_SEGY_NUMCARD):
                header.append(" ".ljust(_SEGY_CARDLEN))
        return header

    def log_txthead(self):
        """Pretty-print a textual header."""
        if self._list is None:
            return
        log.info("%s:", self._info)
        ostr = "\n".join(self._list)
        log.info("%s\n%s", "-------- BEGIN --------", ostr)
        log.info("%s", "--------- END ---------")


if __name__ == "__main__":
    try:
        txt = TxtHeader(encoding="StrangeValue")
    except ValueError:
        log.info("Caught an exception of type ValueError (as expected).")
    try:
        txt = TxtHeader(encoding=42.0)
    except TypeError:
        log.info("Caught an exception of type TypeError (as expected).")
    txt = TxtHeader(encoding="ascii")
    txt = TxtHeader(encoding="AsCiI")
    txt = TxtHeader(encoding="ebcdic")
    myheader = txt.template()
    tmp = list(myheader[0])
    tmp[4:7] = "ABC"
    myheader[0] = "".join(tmp)
    tmp = list(myheader[37])
    tmp[4:7] = "XYZ"
    myheader[37] = "".join(tmp)
    txt.header = myheader
    ebcdic = txt.get_header(decode=False)
    converted = ebcdic.decode("cp037")
    asciihead = txt.get_header(decode=True)
    txt.log_txthead()
    log.info("Length of %s: %d bytes", txt.info, len(txt))
